#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from typing import Optional

from holidays.calendars.gregorian import FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class Spain(HolidayBase, ChristianHolidays, IslamicHolidays, InternationalHolidays):
    """
    References:
     - https://administracion.gob.es/pag_Home/atencionCiudadana/calendarios.html
    """

    country = "ES"
    subdivisions = (
        "AN",
        "AR",
        "AS",
        "CB",
        "CE",
        "CL",
        "CM",
        "CN",
        "CT",
        "EX",
        "GA",
        "IB",
        "MC",
        "MD",
        "ML",
        "NC",
        "PV",
        "RI",
        "VC",
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_holiday(self, *args) -> Optional[date]:
        name, dt = self._parse_holiday(*args)
        if dt.year != self._year:
            return None
        if self.observed and self._is_sunday(dt):
            dt += td(days=+1)
            name = self.tr("%s (Trasladado)") % self.tr(name)
        self[dt] = self.tr(name)
        return dt

    def _populate(self, year):
        super()._populate(year)

        if year != 2023:
            self._add_new_years_day("Año nuevo")

        self._add_epiphany_day("Epifanía del Señor")

        if year >= 2023:
            self._add_holy_thursday("Jueves Santo")

        self._add_good_friday("Viernes Santo")

        if year != 2022:
            self._add_labor_day("Día del Trabajador")

        self._add_assumption_of_mary_day("Asunción de la Virgen")

        self._add_holiday("Día de la Hispanidad", OCT, 12)

        self._add_all_saints_day("Todos los Santos")

        self._add_holiday("Día de la Constitución Española", DEC, 6)

        self._add_immaculate_conception_day("La Inmaculada Concepción")

        if year != 2022:
            self._add_christmas_day("Navidad")

    def _add_subdiv_an_holidays(self):
        if self._year == 2023:
            self._add_new_years_day("Año nuevo")
        self._add_holiday("Día de Andalucia", FEB, 28)
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        if self._year == 2022:
            self._add_labor_day("Día del Trabajador")
            self._add_christmas_day("Navidad")

    def _add_subdiv_ar_holidays(self):
        if self._year == 2023:
            self._add_new_years_day("Año nuevo")
        if self._year <= 2014:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_saint_georges_day("Día de San Jorge")
        if self._year == 2022:
            self._add_labor_day("Día del Trabajador")
            self._add_christmas_day("Navidad")

    def _add_subdiv_as_holidays(self):
        if self._year == 2023:
            self._add_new_years_day("Año nuevo")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Día de Asturias", SEP, 8)
        if self._year == 2022:
            self._add_labor_day("Día del Trabajador")
            self._add_christmas_day("Navidad")

    def _add_subdiv_cb_holidays(self):
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Día de las Instituciones de Cantabria", JUL, 28)
        self._add_holiday("Día de la Bien Aparecida", SEP, 15)
        if self._year == 2022:
            self._add_christmas_day("Navidad")

    def _add_subdiv_ce_holidays(self):
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Nuestra Señora de África", AUG, 5)
        self._add_holiday("Día de la Ciudad Autónoma de Ceuta", SEP, 2)
        if self._year == 2022:
            self._add_eid_al_adha_day("Eid al-Adha")
        elif self._year == 2023:
            self._add_eid_al_adha_day_two("Eid al-Adha")

    def _add_subdiv_cl_holidays(self):
        if self._year == 2023:
            self._add_new_years_day("Año nuevo")
            self._add_saint_james_day("Día de Santiago Apóstol")
        if self._year <= 2014:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Día de Castilla y Leon", APR, 23)
        if self._year == 2022:
            self._add_labor_day("Día del Trabajador")
            self._add_christmas_day("Navidad")

    def _add_subdiv_cm_holidays(self):
        if self._year <= 2015 or 2020 <= self._year <= 2021:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        if self._year <= 2021:
            self._add_easter_monday("Lunes de Pascua")
        if self._year >= 2022:
            self._add_corpus_christi_day("Corpus Christi")
        self._add_holiday("Día de Castilla La Mancha", MAY, 31)
        if self._year == 2022:
            self._add_christmas_day("Navidad")

    def _add_subdiv_cn_holidays(self):
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Día de Canarias", MAY, 30)
        if self._year == 2022:
            self._add_christmas_day("Navidad")

    def _add_subdiv_ct_holidays(self):
        self._add_easter_monday("Lunes de Pascua")
        if self._year == 2022:
            self._add_holiday("Día de la Pascua Granada", JUN, 6)
        self._add_saint_johns_day("San Juan")
        self._add_holiday("Día Nacional de Catalunya", SEP, 11)
        if self._year == 2022:
            self._add_christmas_day("Navidad")
        self._add_christmas_day_two("San Esteban")

    def _add_subdiv_ex_holidays(self):
        if self._year <= 2014:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Día de Extremadura", SEP, 8)
        if self._year == 2023:
            self._add_carnival_tuesday("Carnaval")
        if self._year == 2022:
            self._add_labor_day("Día del Trabajador")
            self._add_christmas_day("Navidad")

    def _add_subdiv_ga_holidays(self):
        if self._year <= 2014 or 2018 <= self._year <= 2021:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        if self._year >= 2022:
            self._add_holiday("Día de las letras Gallegas", MAY, 17)
        if self._year != 2023:
            self._add_saint_johns_day("San Juan")
        self._add_holiday("Día Nacional de Galicia", JUL, 25)

    def _add_subdiv_ib_holidays(self):
        self._add_holiday("Día de las Islas Baleares", MAR, 1)
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_easter_monday("Lunes de Pascua")
        if self._year == 2022:
            self._add_christmas_day("Navidad")
        if self._year <= 2020:
            self._add_christmas_day_two("San Esteban")

    def _add_subdiv_mc_holidays(self):
        if self._year == 2023:
            self._add_new_years_day("Año nuevo")
        if self._year <= 2021 and self._year != 2017:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        if self._year == 2022:
            self._add_labor_day("Día del Trabajador")
        self._add_holiday("Día de la Región de Murcia", JUN, 9)
        if self._year == 2022:
            self._add_christmas_day("Navidad")

    def _add_subdiv_md_holidays(self):
        if self._year <= 2015 or self._year == 2023:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Día de Comunidad de Madrid", MAY, 2)
        if self._year == 2022:
            self._add_saint_james_day("Día de Santiago Apóstol")
            self._add_christmas_day("Navidad")

    def _add_subdiv_ml_holidays(self):
        if self._year <= 2016:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_holiday("Vírgen de la victoria", SEP, 8)
        self._add_holiday("Día de Melilla", SEP, 17)
        if self._year == 2022:
            self._add_eid_al_fitr_day_two("Eid al-Fitr")
            self._add_eid_al_adha_day_three("Eid al-Adha")
            self._add_christmas_day("Navidad")
        elif self._year == 2023:
            self._add_eid_al_fitr_day("Eid al-Fitr")
            self._add_eid_al_adha_day_two("Eid al-Adha")

    def _add_subdiv_nc_holidays(self):
        if self._year <= 2015 or 2018 <= self._year <= 2021:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_easter_monday("Lunes de Pascua")
        if self._year >= 2022:
            self._add_saint_james_day("Día de Santiago Apóstol")
        if self._year == 2022:
            self._add_christmas_day("Navidad")

    def _add_subdiv_pv_holidays(self):
        if self._year <= 2021:
            self._add_saint_josephs_day("San José")
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_easter_monday("Lunes de Pascua")
        if self._year >= 2022:
            self._add_saint_james_day("Día de Santiago Apóstol")
        if self._year <= 2022:
            self._add_holiday("Día de Elcano", SEP, 6)
        if 2011 <= self._year <= 2013:
            self._add_holiday("Día del País Vasco", OCT, 25)

    def _add_subdiv_ri_holidays(self):
        if self._year <= 2022:
            self._add_holy_thursday("Jueves Santo")
        if self._year >= 2022:
            self._add_easter_monday("Lunes de Pascua")
        self._add_holiday("Día de La Rioja", JUN, 9)
        if self._year == 2022:
            self._add_christmas_day("Navidad")

    def _add_subdiv_vc_holidays(self):
        if self._year <= 2022 and self._year != 2017:
            self._add_saint_josephs_day("San José")
        if self._year == 2022:
            self._add_holy_thursday("Jueves Santo")
        self._add_easter_monday("Lunes de Pascua")
        self._add_saint_johns_day("San Juan")
        if self._year <= 2021:
            self._add_holiday("Día de la Comunidad Valenciana", OCT, 9)


class ES(Spain):
    pass


class ESP(Spain):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars import (
    _CustomBuddhistCalendar,
    _CustomChineseCalendar,
    _CustomIslamicCalendar,
    _CustomHinduCalendar,
)
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import (
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
)


class Singapore(
    HolidayBase,
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
):
    country = "SG"
    special_holidays = {
        2001: (NOV, 3, "Polling Day"),
        2006: (MAY, 6, "Polling Day"),
        2011: (MAY, 7, "Polling Day"),
        2015: (
            # SG50 Public holiday
            # Announced on 14 March 2015
            # https://www.mom.gov.sg/newsroom/press-releases/2015/sg50-public-holiday-on-7-august-2015
            (AUG, 7, "SG50 Public Holiday"),
            (SEP, 11, "Polling Day"),
        ),
        2020: (JUL, 10, "Polling Day"),
    }

    def __init__(self, *args, **kwargs) -> None:
        """
        A subclass of :py:class:`HolidayBase` representing public holidays in
        Singapore.

        Limitations:

        - Prior to 1969: holidays are estimated.
        - Prior to 2000: holidays may not be accurate.
        - 2024 and later: the following four moving date holidays (whose exact
          date is announced yearly) are estimated, and so denoted:

          - Hari Raya Puasa
          - Hari Raya Haji
          - Vesak Day
          - Deepavali

        Sources:

        - `Holidays Act <https://sso.agc.gov.sg/Act/HA1998>`__ (Act 24 of
          1968—Holidays (Amendment) Act 1968)
        - `Ministry of Manpower
          <https://www.mom.gov.sg/employment-practices/public-holidays>`__

        References:

        - `Wikipedia
          <https://en.wikipedia.org/wiki/Public_holidays_in_Singapore>`__

        Country created and maintained by: `Mike Borsetti
        <https://github.com/mborsetti>`__

        See parameters and usage in :py:class:`HolidayBase`.
        """
        BuddhistCalendarHolidays.__init__(
            self, calendar=SingaporeBuddhistCalendar(), show_estimated=True
        )
        ChineseCalendarHolidays.__init__(
            self, calendar=SingaporeChineseCalendar(), show_estimated=True
        )
        ChristianHolidays.__init__(self)
        HinduCalendarHolidays.__init__(self, calendar=SingaporeHinduCalendar())
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=SingaporeIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _populate(self, year) -> None:
        super()._populate(year)
        observed_dates = set()

        # New Year's Day
        observed_dates.add(self._add_new_years_day("New Year's Day"))

        # Chinese New Year (two days)
        name = "Chinese New Year"
        observed_dates.add(self._add_chinese_new_years_day(name))  # type: ignore[arg-type]
        observed_dates.add(self._add_chinese_new_years_day_two(name))  # type: ignore[arg-type]

        # Hari Raya Puasa (Eid al-Fitr)
        observed_dates.update(self._add_eid_al_fitr_day("Hari Raya Puasa"))
        if year <= 1968:
            self._add_eid_al_fitr_day_two("Second day of Hari Raya Puasa")

        # Hari Raya Haji (Eid al-Adha)
        observed_dates.update(self._add_eid_al_adha_day("Hari Raya Haji"))

        # Good Friday
        self._add_good_friday("Good Friday")

        if year <= 1968:
            # Holy Saturday
            self._add_holy_saturday("Holy Saturday")

            # Easter Monday
            self._add_easter_monday("Easter Monday")

        # Labour Day
        observed_dates.add(self._add_labor_day("Labour Day"))

        # Vesak Day
        observed_dates.add(self._add_vesak("Vesak Day"))  # type: ignore[arg-type]

        # National Day
        observed_dates.add(self._add_holiday("National Day", AUG, 9))  # type: ignore[arg-type]

        # Deepavali (Diwali)
        observed_dates.add(self._add_diwali("Deepavali"))  # type: ignore[arg-type]

        # Christmas Day
        observed_dates.add(self._add_christmas_day("Christmas Day"))

        # Boxing day (up to and including 1968)
        if year <= 1968:
            self._add_christmas_day_two("Boxing Day")

        # Implement Section 4(2) of the Holidays Act:
        # "if any day specified in the Schedule falls on a Sunday,
        # the day next following not being itself a public holiday
        # is declared a public holiday in Singapore."
        if not self.observed:
            return None
        if year >= 1998:
            for dt in sorted(observed_dates):
                if not self._is_sunday(dt):
                    continue
                self._add_holiday(
                    "%s (Observed)" % self[dt],
                    dt + td(days=2 if dt + td(days=+1) in observed_dates else 1),
                )

        # special case (observed from previous year)
        if year == 2007:
            self._add_holiday("Hari Raya Haji (Observed)", JAN, 2)


class SG(Singapore):
    pass


class SGP(Singapore):
    pass


class SingaporeBuddhistCalendar(_CustomBuddhistCalendar):
    VESAK_DATES = {
        2001: (MAY, 7),
        2002: (MAY, 26),
        2003: (MAY, 15),
        2004: (JUN, 2),
        2005: (MAY, 22),
        2006: (MAY, 12),
        2007: (MAY, 31),
        2008: (MAY, 19),
        2009: (MAY, 9),
        2010: (MAY, 28),
        2011: (MAY, 17),
        2012: (MAY, 5),
        2013: (MAY, 24),
        2014: (MAY, 13),
        2015: (JUN, 1),
        2016: (MAY, 21),
        2017: (MAY, 10),
        2018: (MAY, 29),
        2019: (MAY, 19),
        2020: (MAY, 7),
        2021: (MAY, 26),
        2022: (MAY, 15),
        2023: (JUN, 2),
    }


class SingaporeChineseCalendar(_CustomChineseCalendar):
    LUNAR_NEW_YEAR_DATES = {
        2001: (JAN, 24),
        2002: (FEB, 12),
        2003: (FEB, 1),
        2004: (JAN, 22),
        2005: (FEB, 9),
        2006: (JAN, 30),
        2007: (FEB, 19),
        2008: (FEB, 7),
        2009: (JAN, 26),
        2010: (FEB, 14),
        2011: (FEB, 3),
        2012: (JAN, 23),
        2013: (FEB, 10),
        2014: (JAN, 31),
        2015: (FEB, 19),
        2016: (FEB, 8),
        2017: (JAN, 28),
        2018: (FEB, 16),
        2019: (FEB, 5),
        2020: (JAN, 25),
        2021: (FEB, 12),
        2022: (FEB, 1),
        2023: (JAN, 22),
    }


class SingaporeHinduCalendar(_CustomHinduCalendar):
    DIWALI_DATES = {
        2001: (NOV, 14),
        2002: (NOV, 3),
        2003: (OCT, 23),
        2004: (NOV, 11),
        2005: (NOV, 1),
        2006: (OCT, 21),
        2007: (NOV, 8),
        2008: (OCT, 27),
        2009: (NOV, 15),
        2010: (NOV, 5),
        2011: (OCT, 26),
        2012: (NOV, 13),
        2013: (NOV, 2),
        2014: (OCT, 22),
        2015: (NOV, 10),
        2016: (OCT, 29),
        2017: (OCT, 18),
        2018: (NOV, 6),
        2019: (OCT, 27),
        2020: (NOV, 14),
        2021: (NOV, 4),
        2022: (OCT, 24),
        2023: (NOV, 12),
    }


class SingaporeIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2004: (FEB, 1),
        2005: (JAN, 21),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 8),
        2009: (NOV, 27),
        2010: (NOV, 17),
        2011: (NOV, 6),
        2012: (OCT, 26),
        2013: (OCT, 15),
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 12),
        2017: (SEP, 1),
        2018: (AUG, 22),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 10),
        2023: (JUN, 29),
    }

    EID_AL_FITR_DATES = {
        2001: (DEC, 16),
        2002: (DEC, 6),
        2003: (NOV, 25),
        2004: (NOV, 14),
        2005: (NOV, 3),
        2006: (OCT, 24),
        2007: (OCT, 13),
        2008: (OCT, 1),
        2009: (SEP, 20),
        2010: (SEP, 10),
        2011: (AUG, 30),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 28),
        2015: (JUL, 17),
        2016: (JUL, 6),
        2017: (JUN, 25),
        2018: (JUN, 15),
        2019: (JUN, 5),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 3),
        2023: (APR, 22),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars import _CustomIslamicCalendar
from holidays.calendars.gregorian import JAN, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays, IslamicHolidays


class Chad(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Chad
      - https://www.ilo.org/dyn/natlex/docs/ELECTRONIC/97323/115433/F-316075167/TCD-97323.pdf
    """

    country = "TD"
    special_holidays = {
        2021: (APR, 23, "Funeral of Idriss Déby Itno"),
    }

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=ChadIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date) -> None:
        if self.observed and self._is_sunday(dt):
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=+1))

    def _populate(self, year):
        # On 11 August 1960, Chad gained independence from France.
        if year <= 1960:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_observed(self._add_new_years_day("New Year's Day"))

        # International Women's Day.
        self._add_observed(self._add_womens_day("International Women's Day"))

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Labour Day.
        self._add_observed(self._add_labor_day("Labour Day"))

        # Independence Day.
        self._add_observed(self._add_holiday("Independence Day", AUG, 11))

        # All Saints' Day.
        self._add_all_saints_day("All Saints' Day")

        # Republic Day.
        self._add_observed(self._add_holiday("Republic Day", NOV, 28))

        if year >= 1991:
            # Freedom and Democracy Day.
            self._add_observed(self._add_holiday("Freedom and Democracy Day", DEC, 1))

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Eid al-Fitr.
        self._add_eid_al_fitr_day("Eid al-Fitr")

        # Eid al-Adha.
        self._add_eid_al_adha_day("Eid al-Adha")

        # Mawlid.
        self._add_mawlid_day("Mawlid")


class TD(Chad):
    pass


class TCD(Chad):
    pass


class ChadIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2015: (SEP, 24),
        2016: (SEP, 13),
        2017: (SEP, 2),
        2018: (AUG, 22),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 9),
        2023: (JUN, 28),
    }

    EID_AL_FITR_DATES = {
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2018: (JUN, 15),
        2019: (JUN, 4),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 21),
    }

    MAWLID_DATES = {
        2015: ((JAN, 3), (DEC, 24)),
        2016: (DEC, 12),
        2017: (DEC, 1),
        2018: (NOV, 21),
        2019: (NOV, 9),
        2020: (OCT, 29),
        2021: (OCT, 18),
        2022: (OCT, 8),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import InternationalHolidays, ThaiCalendarHolidays


class Thailand(HolidayBase, InternationalHolidays, ThaiCalendarHolidays):
    """
    A subclass of :py:class:`HolidayBase` representing public holidays in Thailand.

    References:

    - Based on: https://en.wikipedia.org/wiki/Public_holidays_in_Thailand
    - Checked with: (Bank of Thailand 's wbm) http://tiny.cc/wa_bot_2023
    - [In Lieus]
        (isranews.org 's wbm) http://tiny.cc/wa_isranews_inlieu_hist
        https://resolution.soc.go.th/?prep_id=99159317
        https://resolution.soc.go.th/?prep_id=196007
        https://github.com/dr-prodigy/python-holidays/pull/929
    - [New Year's Day]
        (wikisource.org 's wbm) http://tiny.cc/wa_wiki_thai_newyear_2483
    - [Chakri Memorial Day]
        (ocac.got.th 's wbm) http://tiny.cc/wa_ocac_chakri
    - [Songkran Festival]
        (museumsiam.org 's wbm) http://tiny.cc/wa_museumsiam_songkran
        https://resolution.soc.go.th/?prep_id=123659
    - [National Labour Day]
        https://www.thairath.co.th/lifestyle/culture/1832869
    - [National Day (24 June: Defunct)]
        (Ministry of Culture 's wbm) http://tiny.cc/wa_mincul_nat_day
    - [Coronation Day]
        https://www.matichon.co.th/politics/news_526200
        https://workpointtoday.com/news1-5/
    - [HM Queen Suthida's Birthday]
        https://www.thairath.co.th/news/politic/1567418
    - [HM Maha Vajiralongkorn's Birthday]
        https://www.matichon.co.th/politics/news_526200
    - [HM Queen Sirikit the Queen Mother's Birthday]
        https://hilight.kapook.com/view/14164
    - [National Mother's Day]
        https://www.brh.go.th/index.php/2019-02-27-04-11-52/542-12-2564
    - [HM King Bhumibol Adulyadej Memorial Day]
        https://www.matichon.co.th/politics/news_526200
    - [HM King Chulalongkorn Memorial Day]
        https://th.wikipedia.org/wiki/วันปิยมหาราช
    - [HM King Bhumibol Adulyadej's Birthday]
        (Ministry of Culture 's wbm) http://tiny.cc/wa_mincul_nat_day
        https://hilight.kapook.com/view/148862
    - [National Father's Day]
        https://www.brh.go.th/index.php/2019-02-27-04-12-21/594-5-5
    - [Constitution Day]
        https://hilight.kapook.com/view/18208
        (Bank of Thailand 's wbm) http://tiny.cc/wa_bot_1992
        https://www.myhora.com/ปฏิทิน/ปฏิทิน-พ.ศ.2475.aspx
    - [New Year's Eve]
        (Bank of Thailand 's wbm) http://tiny.cc/wa_bot_1992
        https://resolution.soc.go.th/?prep_id=205799
        https://resolution.soc.go.th/?prep_id=210744
    - [Makha Bucha]
        https://www.onab.go.th/th/content/category/detail/id/73/iid/3403
    - [Visakha Bucha]
        https://www.onab.go.th/th/content/category/detail/id/73/iid/3401
    - [Asarnha Bucha]
        https://www.onab.go.th/th/content/category/detail/id/73/iid/3397
    - [Buddhist Lent Day]
        https://www.onab.go.th/th/content/category/detail/id/73/iid/3395
    - [Royal Ploughing Ceremony]
        https://en.wikipedia.org/wiki/Royal_Ploughing_Ceremony
        https://www.lib.ru.ac.th/journal/may/may_phauchmongkol.html
        https://www.myhora.com/ปฏิทิน/ปฏิทิน-พ.ศ.2540.aspx


    Limitations:

    - This is only 100% accurate for 1997-2023; any future dates are up to the
      Royal Thai Government Gazette which updates on a year-by-year basis.

    - Approx. date only goes as far back as 1941 (B.E. 2484) as the Thai
      calendar for B.E. 2483 as we only have nine months from switching
      New Year Date (April 1st to January 1st).

    - Thai Lunar Calendar Holidays only work from 1941 (B.E. 2484) onwards
      until 2057 (B.E. 2600) as we only have Thai year-type data for
      cross-checking until then.

    - Royal Ploughing Ceremony Day is date is announced on an annual basis
      by the Court Astrologers, thus need an annual update to the library here

    - This doesn't cover Thai regional public holidays yet, only stubs added


    Country created by: `arkid15r <https://github.com/arkid15r>`__

    Country maintained by: `PPsyrius <https://github.com/PPsyrius>`__
    """

    country = "TH"
    default_language = "th"

    # วันหยุดพิเศษ (เพิ่มเติม) - see Bank of Thailand's DB for Cross-Check.

    # Special In Lieu Holiday.
    thai_special_in_lieu_holidays = tr("วันหยุดชดเชย")
    # Thai Election Day.
    thai_election = tr("วันเลือกตั้ง")
    # Thai Election Day (in lieu).
    thai_election_in_lieu = tr("ชดเชยวันเลือกตั้ง")
    # Bridge Public Holiday.
    thai_bridge_public_holiday = tr("วันหยุดพิเศษ (เพิ่มเติม)")

    # Special Cases.

    # HM King Bhumibol Adulyadej's Golden Jubilee.
    rama_ix_golden_jubilee = tr("พระราชพิธีกาญจนาภิเษก พ.ศ. 2539")
    # HM King Bhumibol Adulyadej's 60th Anniversary of Accession Event.
    rama_ix_sixty_accession = tr("พระราชพิธีฉลองสิริราชสมบัติครบ 60 ปี พ.ศ. 2549")
    # Emergency Lockdown (Thai Military Coup d'état).
    thai_military_emergency_lockdown = tr("วันหยุดพิเศษ (คมช.)")
    # Emergency Lockdown (Thai Political Unrest).
    thai_political_emergency_lockdown = tr("วันหยุดพิเศษ (การเมือง)")
    # Emergency Lockdown (2011 Thailand Floods).
    thai_flood_2011_emergency_lockdown = tr("วันหยุดพิเศษ (มหาอุทกภัย พ.ศ. 2554)")
    # Day of Mourning for HM King Bhumibol Adulyadej.
    rama_ix_mourning = tr("วันหยุดพิเศษ (ร่วมถวายอาลัย ส่งดวงพระวิญญาณพระบรมศพ)")
    # HM King Bhumibol Adulyadej's Royal Cremation Ceremony.
    rama_ix_cremation = tr(
        "วันพระราชพิธีถวายพระเพลิงพระบรมศพพระบาทสมเด็จพระปรมินทรมหาภูมิพลอดุลยเดช"
    )
    # HM King Maha Vajiralongkorn's Coronation Celebrations.
    rama_x_coronation_celebrations = tr(
        "พระราชพิธีบรมราชาภิเษก พระบาทสมเด็จพระวชิรเกล้าเจ้าอยู่หัว"
    )
    # Songkran Festival (in lieu).
    songkran_festival_in_lieu_covid = tr("ชดเชยวันสงกรานต์")

    special_holidays = {
        # 1992-1994 (include In Lieus, Checked with Bank of Thailand Data).
        1992: (
            (MAY, 18, thai_special_in_lieu_holidays),
            (DEC, 7, thai_special_in_lieu_holidays),
        ),
        1993: (
            (MAR, 8, thai_special_in_lieu_holidays),
            (MAY, 3, thai_special_in_lieu_holidays),
            (OCT, 25, thai_special_in_lieu_holidays),
            (DEC, 6, thai_special_in_lieu_holidays),
        ),
        1994: (
            (JAN, 3, thai_special_in_lieu_holidays),
            (MAY, 2, thai_special_in_lieu_holidays),
            (JUL, 25, thai_special_in_lieu_holidays),
            (OCT, 24, thai_special_in_lieu_holidays),
            (DEC, 12, thai_special_in_lieu_holidays),
        ),
        # 1995-1997 (Bank of Thailand Data).
        1996: (JUN, 10, rama_ix_golden_jubilee),
        # 1998-2000 (include In Lieus, Checked with Bank of Thailand Data).
        1998: (
            (MAY, 11, thai_special_in_lieu_holidays),
            (DEC, 7, thai_special_in_lieu_holidays),
        ),
        1999: (
            (MAY, 3, thai_special_in_lieu_holidays),
            (MAY, 31, thai_special_in_lieu_holidays),
            (OCT, 25, thai_special_in_lieu_holidays),
            (DEC, 6, thai_special_in_lieu_holidays),
        ),
        2000: (
            (JAN, 3, thai_special_in_lieu_holidays),
            (FEB, 21, thai_special_in_lieu_holidays),
            (AUG, 14, thai_special_in_lieu_holidays),
            (DEC, 11, thai_special_in_lieu_holidays),
            (DEC, 29, thai_election),
        ),
        # From 2001 Onwards (Checked with Bank of Thailand Data).
        2006: (
            (APR, 19, thai_election),
            (JUN, 9, rama_ix_sixty_accession),
            (JUN, 12, rama_ix_sixty_accession),
            (JUN, 13, rama_ix_sixty_accession),
            (SEP, 20, thai_military_emergency_lockdown),
        ),
        2007: (DEC, 24, thai_election_in_lieu),
        2009: (
            (JAN, 2, thai_bridge_public_holiday),
            (APR, 10, thai_political_emergency_lockdown),
            (APR, 16, thai_political_emergency_lockdown),
            (APR, 17, thai_political_emergency_lockdown),
            (JUL, 6, thai_bridge_public_holiday),
        ),
        2010: (
            (MAY, 20, thai_bridge_public_holiday),
            (MAY, 21, thai_bridge_public_holiday),
            (AUG, 13, thai_bridge_public_holiday),
        ),
        2011: (
            (MAY, 16, thai_bridge_public_holiday),
            (OCT, 27, thai_flood_2011_emergency_lockdown),
            (OCT, 28, thai_flood_2011_emergency_lockdown),
            (OCT, 29, thai_flood_2011_emergency_lockdown),
            (OCT, 30, thai_flood_2011_emergency_lockdown),
            (OCT, 31, thai_flood_2011_emergency_lockdown),
        ),
        2012: (APR, 9, thai_bridge_public_holiday),
        2013: (DEC, 30, thai_bridge_public_holiday),
        2014: (AUG, 11, thai_bridge_public_holiday),
        2015: (
            (JAN, 2, thai_bridge_public_holiday),
            (MAY, 4, thai_bridge_public_holiday),
        ),
        2016: (
            (MAY, 6, thai_bridge_public_holiday),
            (JUL, 18, thai_bridge_public_holiday),
            (OCT, 14, rama_ix_mourning),
        ),
        2017: (OCT, 26, rama_ix_cremation),
        2019: (MAY, 6, rama_x_coronation_celebrations),
        2020: (
            (JUL, 27, songkran_festival_in_lieu_covid),
            (SEP, 4, songkran_festival_in_lieu_covid),
            (SEP, 7, songkran_festival_in_lieu_covid),
            (NOV, 19, thai_bridge_public_holiday),
            (NOV, 20, thai_bridge_public_holiday),
            (DEC, 11, thai_bridge_public_holiday),
        ),
        2021: (
            (FEB, 12, thai_bridge_public_holiday),
            (APR, 12, thai_bridge_public_holiday),
            (SEP, 24, thai_bridge_public_holiday),
        ),
        2022: (
            (JUL, 15, thai_bridge_public_holiday),
            (JUL, 29, thai_bridge_public_holiday),
            (OCT, 14, thai_bridge_public_holiday),
            (DEC, 30, thai_bridge_public_holiday),
        ),
        2023: (MAY, 5, thai_bridge_public_holiday),
    }
    supported_languages = ("en_US", "th")

    def __init__(self, **kwargs) -> None:
        InternationalHolidays.__init__(self)
        ThaiCalendarHolidays.__init__(self)
        super().__init__(**kwargs)

    def _populate(self, year):
        # Due to Thai Calendar Migration, this is capped off at 1941.
        # But certain holidays were implemented before 1941.
        if year <= 1940:
            return None

        def _add_observed(dt: date) -> None:
            """
            !!! If Public Holiday falls on weekends, (in lieu) on workday !!!
            Despite the wording, this usually only applies to Monday only for
            holidays, consecutive holidays all have their own special in lieu
            declared separately.

            Data from 1992-1994 and 1998-2000 are declared discretely in
            special_holidays declarations above.

            Applied Automatically for Monday if on Weekends: 1961-1973
             **NOTE: No New Year's Eve (in lieu) for this period
            No In Lieu days available: 1974-1988
            Case-by-Case application for Workday if on Weekends: 1989-1994
            Applied Automatically for Workday if on Weekends: 1995-1997
            Case-by-Case application for Workday if on Weekends: 1998-2000
            Applied Automatically for Workday if on Weekends: 2001-Present
            """
            if (
                self.observed
                and self._is_weekend(dt)
                and (1961 <= year <= 1973 or 1995 <= year <= 1997 or year >= 2001)
            ):
                in_lieu = dt + td(days=+2 if self._is_saturday(dt) else +1)
                for name in self.get_list(dt):
                    # %s (in lieu)
                    self._add_holiday(self.tr("ชดเชย%s") % name, in_lieu)

        super()._populate(year)

        # Fixed Date Holidays

        # วันขึ้นปีใหม่
        # Status: In-Use.
        # Starts in the present form in 1941 (B.E. 2484).
        # TODO: Add check for 1941 if we support earlier dates.

        # New Year's Day.
        jan_1 = self._add_new_years_day(tr("วันขึ้นปีใหม่"))
        _add_observed(jan_1)

        # วันหยุดชดเชยวันสิ้นปี
        # Status: In-Use.
        # Added separately from New Year's Eve itself so that it would't
        #   go over the next year.
        #   - CASE 1: SAT-SUN -> 1 in-lieu on TUE.
        #   - CASE 2: SUN-MON -> 1 in-lieu on TUE.
        # See in lieu logic in `_add_observed(dt: date)`.

        new_years_eve_in_lieu = self.tr("ชดเชย%s") % self.tr("วันสิ้นปี")

        if self.observed and (1995 <= year <= 1997 or year >= 2001):
            if self._is_sunday(jan_1):
                self._add_new_years_day_three(new_years_eve_in_lieu)
            elif self._is_monday(jan_1):
                self._add_new_years_day_two(new_years_eve_in_lieu)

        # วันจักรี
        # Status: In-Use.
        # Starts in present form in 1918 (B.E. 2461).
        # TODO: Add check for 1918 if we support earlier dates.

        # Chakri Memorial Day.
        _add_observed(self._add_holiday(tr("วันจักรี"), APR, 6))

        # วันสงกรานต์
        # Status: In-Use.
        # Used to be April 1st as Thai New Year Day.
        # Initially abandoned in 1941 (B.E. 2484), declared again as
        #   public holidays in 1948 (2491 B.E).
        #  - 1948-1953, celebrated on Apr 13-15
        #  - 1954-1956, abandoned as a public holiday
        #  - 1957-1988, only celebrated on Apr 13
        #  - 1989-1997, celebrated on Apr 12-14
        #  - 1998-Present, celebrated on Apr 13-15
        #    (Except for 2020 due to Covid-19 outbreaks)
        # This has its own in-lieu trigger.

        if year >= 1948:
            # Songkran Festival.
            songkran_festival = tr("วันสงกรานต์")
            if year <= 1953 or (1957 <= year != 2020):
                dt = self._add_holiday(songkran_festival, APR, 12 if 1989 <= year <= 1997 else 13)
                if 1957 <= year <= 1988:
                    _add_observed(dt)
                else:
                    self._add_holiday(songkran_festival, dt + td(days=+1))
                    self._add_holiday(songkran_festival, dt + td(days=+2))

            # วันหยุดชดเชยวันสงกรานต์
            # If Songkran happened to be held on the weekends, only one in-lieu
            #   public holiday is added, No in lieus for SUN-MON-TUE case.
            #   - CASE 1: THU-FRI-SAT -> 1 in-lieu on MON
            #   - CASE 2: FRI-SAT-SUN -> 1 in-lieu on MON
            #   - CASE 3: SAT-SUN-MON -> 1 in-lieu on TUE
            # See in lieu logic in `_add_observed(dt: date)`.
            # Status: In Use.

            songkran_festival_in_lieu = self.tr("ชดเชย%s") % songkran_festival

            if self.observed and (1995 <= year <= 1997 or 2001 <= year != 2020):
                if self._is_thursday(dt):
                    self._add_holiday(songkran_festival_in_lieu, dt + td(days=+4))
                elif self._is_friday(dt) or self._is_saturday(dt):
                    self._add_holiday(songkran_festival_in_lieu, dt + td(days=+3))

        # วันแรงงานแห่งชาติ
        # Status: In-Use.
        # Starts in the present form in 1974 (B.E. 2517).
        # Does existed officially since 1956 (B.E. 2499), but wasn't a public holiday until then.
        # *** NOTE: only observed by financial and private sectors.

        if year >= 1974:
            # National Labour day.
            _add_observed(self._add_labor_day(tr("วันแรงงานแห่งชาติ")))

        # วันชาติ
        # Status: In-Use.
        # Starts in 1939 (B.E. 2482) by Plaek Phibunsongkhram.
        # Replaced by Rama IX's birthday in 1960 (B.E. 2503) by Sarit Thanarat.
        # TODO: Add check for 1939 if we support earlier dates.

        # National Day.
        _add_observed(self._add_holiday(tr("วันชาติ"), *((JUN, 24) if year <= 1959 else (DEC, 5))))

        # วันฉัตรมงคล
        # Starts in 1958 (B.E. 2501) for Rama IX's Coronation: May 5th.
        # No celebration in 2017-2019 (B.E 2560-2562).
        # Reestablished with Rama X's Coronation in 2020: May 4th.

        # Coronation Day.
        coronation_day = tr("วันฉัตรมงคล")

        if 1958 <= year <= 2016:
            _add_observed(self._add_holiday(coronation_day, MAY, 5))
        elif year >= 2020:
            _add_observed(self._add_holiday(coronation_day, MAY, 4))

        # วันเฉลิมพระชนมพรรษา พระราชินี
        # Status: In-Use.
        # Starts in 2019 (B.E. 2562).

        if year >= 2019:
            _add_observed(
                self._add_holiday(
                    # HM Queen Suthida's Birthday.
                    tr("วันเฉลิมพระชนมพรรษาสมเด็จพระนางเจ้าสุทิดา พัชรสุธาพิมลลักษณ พระบรมราชินี"),
                    JUN,
                    3,
                )
            )

        # วันเฉลิมพระชนมพรรษา รัชกาลที่ 10
        # Status: In-Use.
        # Started in 2017 (B.E 2560).

        if year >= 2017:
            _add_observed(
                self._add_holiday(
                    # HM King Maha Vajiralongkorn's Birthday.
                    tr(
                        "วันเฉลิมพระชนมพรรษาพระบาทสมเด็จพระปรเมนทรรามาธิบดี"
                        "ศรีสินทรมหาวชิราลงกรณ พระวชิรเกล้าเจ้าอยู่หัว"
                    ),
                    JUL,
                    28,
                )
            )

        # วันเฉลิมพระชนมพรรษา พระบรมราชินีนาถ ( 1976-2017)
        # วันเฉลิมพระชนมพรรษา พระบรมราชชนนีพันปีหลวง (2017-Present)
        # Status: In-Use.
        # Started in 1976 (B.E. 2519) alongside Mother's Day.
        # Initial celebration as HM Queen Sirikit's Birthday.
        # Now acts as the Queen Mother from 2017 onwards.

        if year >= 1976:
            name = (
                # HM Queen Sirikit the Queen Mother's Birthday.
                tr("วันเฉลิมพระชนมพรรษาสมเด็จพระบรมราชชนนีพันปีหลวง")
                if year >= 2017
                # HM Queen Sirikit's Birthday.
                else tr("วันเฉลิมพระชนมพรรษาสมเด็จพระนางเจ้าสิริกิติ์ พระบรมราชินีนาถ")
            )
            _add_observed(self._add_holiday(name, AUG, 12))

        # วันแม่แห่งชาติ
        # Status: In-Use.
        # Started 1950 (B.E 2493) initially as April 15 and cancelled in
        #   1958 (B.E 2501) when the Min. of Culture was abolished.
        # Restarts again in 1976 (B.E. 2519) on Queen Sirikit's Birthday
        #   (August 12) and stay that way from that point onwards.

        # National Mother's Day.
        thai_mothers_day = tr("วันแม่แห่งชาติ")

        if 1950 <= year <= 1957:
            _add_observed(self._add_holiday(thai_mothers_day, APR, 15))
        elif year >= 1976:
            _add_observed(self._add_holiday(thai_mothers_day, AUG, 12))

        # วันคล้ายวันสวรรคตพระบาทสมเด็จพระปรมินทร มหาภูมิพลอดุลยเดช บรมนาถบพิตร
        # Status: In-Use.
        # Started in 2017 (B.E 2560).
        # Got conferred with 'the Great' title in 2019 (B.E. 2562).

        if year >= 2017:
            name = (
                # Anniversary for the Death of King Bhumibol Adulyadej the Great.
                tr(
                    "วันคล้ายวันสวรรคตพระบาทสมเด็จพระบรมชนกาธิเบศร "
                    "มหาภูมิพลอดุลยเดชมหาราช บรมนาถบพิตร"
                )
                if year >= 2019
                # Anniversary for the Death of King Bhumibol Adulyadej.
                else tr("วันคล้ายวันสวรรคตพระบาทสมเด็จพระปรมินทรมหาภูมิพลอดุลยเดช บรมนาถบพิตร")
            )
            _add_observed(self._add_holiday(name, OCT, 13))

        # วันปิยมหาราช
        # Status: In-Use.
        # Started in 1911 (B.E. 2454).
        # TODO: Add check for 1911 if we support earlier dates.

        # HM King Chulalongkorn Memorial Day.
        _add_observed(self._add_holiday(tr("วันปิยมหาราช"), OCT, 23))

        # วันเฉลิมพระชนมพรรษา รัชกาลที่ 9 (1960-2016)
        # วันคล้ายวันเฉลิมพระชนมพรรษา รัชกาลที่ 9 (2017-Present)
        # Status: In-Use.
        # Replaced Nataion Day (26 June) in 1960 (B.E. 2503) by Sarit Thanarat.
        # Confirmed as still in-use in 2017.
        # Got conferred with 'the Great' title in 2019 (B.E. 2562).

        if year >= 1960:
            name = (
                # HM King Bhumibol Adulyadej's the Great's Birthday Anniversary.
                tr(
                    "วันคล้ายวันเฉลิมพระชนมพรรษาพระบาทสมเด็จพระบรม"
                    "ชนกาธิเบศร มหาภูมิพลอดุลยเดชมหาราช บรมนาถบพิตร"
                )
                if year >= 2019
                else (
                    # HM King Bhumibol Adulyadej Birthday Anniversary.
                    tr(
                        "วันคล้ายวันเฉลิมพระชนมพรรษา"
                        "พระบาทสมเด็จพระปรมินทรมหาภูมิพลอดุลยเดช บรมนาถบพิตร"
                    )
                    if year >= 2016
                    # HM King Bhumibol Adulyadej Birthday Anniversary.
                    else tr(
                        "วันเฉลิมพระชนมพรรษาพระบาทสมเด็จพระปรมินทรมหาภูมิพลอดุลยเดช บรมนาถบพิตร"
                    )
                )
            )
            _add_observed(self._add_holiday(name, DEC, 5))

        # วันพ่อแห่งชาติ
        # Status: In-Use.
        # Starts in 1980 (B.E 2523).
        # Technically, a replication of HM King Bhumibol Adulyadej's Birthday
        #   but it's in the official calendar, so may as well have this here.

        if year >= 1980:
            # National Father's Day.
            _add_observed(self._add_holiday(tr("วันพ่อแห่งชาติ"), DEC, 5))

        # วันรัฐธรรมนูญ
        # Status: In-Use.
        # Presumed to starts in 1932 (B.E. 2475).
        # Last known official record is Bank of Thailand's in 1992 (B.E. 2535).
        # TODO: Add check for 1932 if we support earlier dates.

        # Constitution Day.
        _add_observed(self._add_holiday(tr("วันรัฐธรรมนูญ"), DEC, 10))

        # วันสิ้นปี
        # Status: In-Use.
        # Presumed to start in the present form in 1941 (B.E. 2484).
        # Last known official record is Bank of Thailand's in 1992 (B.E. 2535).
        # TODO: Add check for 1941 if we support earlier dates.
        # This has its own in-lieu trigger.

        # New Year's Eve.
        self._add_new_years_eve(tr("วันสิ้นปี"))

        # Thai Lunar Calendar Holidays
        # See `_ThaiLunisolar` in holidays/utils.py for more details.
        # Thai Lunar Calendar Holidays only work from 1941 to 2057.

        # วันมาฆบูชา
        # Status: In-Use.

        # Makha Bucha.
        makha_bucha_date = self._add_makha_bucha(tr("วันมาฆบูชา"))
        if makha_bucha_date:
            _add_observed(makha_bucha_date)

        # วันวิสาขบูชา
        # Status: In-Use.

        # Visakha Bucha.
        visakha_bucha_date = self._add_visakha_bucha(tr("วันวิสาขบูชา"))
        if visakha_bucha_date:
            _add_observed(visakha_bucha_date)

        # วันอาสาฬหบูชา
        # Status: In-Use.
        # This has its own in-lieu trigger.

        # Asarnha Bucha.
        asarnha_bucha_date = self._add_asarnha_bucha(tr("วันอาสาฬหบูชา"))

        # วันเข้าพรรษา
        # Status: In-Use.
        # This has its own in-lieu trigger.

        # Buddhist Lent Day.
        self._add_khao_phansa(tr("วันเข้าพรรษา"))

        # วันหยุดชดเชยวันอาสาฬหบูชา
        # วันหยุดชดเชยวันเข้าพรรษา
        # Status: In Use.
        #  - CASE 1: FRI-SAT -> 1 in-lieu on MON
        #  - CASE 2: SAT-SUN -> 1 in-lieu on MON
        #  - CASE 3: SUN-MON -> 1 in-lieu on TUE
        # See in lieu logic in `_add_observed(dt: date)`.

        if (
            asarnha_bucha_date
            and self.observed
            and (1961 <= year <= 1973 or 1995 <= year <= 1997 or year >= 2001)
        ):
            if self._is_friday(asarnha_bucha_date):
                self._add_holiday(
                    self.tr("ชดเชย%s") % self.tr("วันเข้าพรรษา"), asarnha_bucha_date + td(days=+3)
                )
            elif self._is_weekend(asarnha_bucha_date):
                self._add_holiday(
                    self.tr("ชดเชย%s") % self.tr("วันอาสาฬหบูชา"), asarnha_bucha_date + td(days=+2)
                )

        # No Future Fixed Date Holidays

        # วันพืชมงคล
        # Restarts in 1957 (B.E. 2500).
        # Is dated on an annual basis by the Royal Palace.
        # This isn't even fixed even by the Thai Lunar Calendar, but instead
        #   by Court Astrologers; All chosen dates are all around May, so we
        #   can technically assign it to 13 May for years prior with no data.
        # *** NOTE: only observed by government sectors.
        # TODO: Update this annually around Dec of each year.

        # Royal Ploughing Ceremony.
        raeknakhwan = tr("วันพืชมงคล")

        raeknakhwan_dates = {
            1997: (MAY, 13),
            1998: (MAY, 13),
            # Not held in 1999 date.
            2000: (MAY, 15),
            2001: (MAY, 16),
            2002: (MAY, 9),
            2003: (MAY, 8),
            2004: (MAY, 7),
            2005: (MAY, 11),
            2006: (MAY, 11),
            2007: (MAY, 10),
            2008: (MAY, 9),
            2009: (MAY, 11),
            2010: (MAY, 10),
            2011: (MAY, 13),
            2012: (MAY, 9),
            2013: (MAY, 13),
            2014: (MAY, 9),
            2015: (MAY, 13),
            2016: (MAY, 9),
            2017: (MAY, 12),
            2018: (MAY, 14),
            2019: (MAY, 9),
            2020: (MAY, 11),
            2021: (MAY, 13),
            2022: (MAY, 17),
            2023: (MAY, 11),
        }
        # For years with exact date data.
        if year in raeknakhwan_dates:
            _add_observed(self._add_holiday(raeknakhwan, *raeknakhwan_dates[year]))
        # Approx. otherwise for 1957-2013.
        elif 1957 <= year <= 1996:
            _add_observed(self._add_holiday(raeknakhwan, MAY, 13))


class TH(Thailand):
    pass


class THA(Thailand):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import APR, MAY, JUL, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Curacao(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://loketdigital.gobiernu.cw/Loket/product/571960bbe1e5fe8712b10a1323630e70
    https://en.wikipedia.org/wiki/Public_holidays_in_Cura%C3%A7ao

    New Year's Eve (Vispu di Aña Nobo) is a half-day public holiday, though
    this isn't supported by Python Holidays so it won't be implemented.
    """

    country = "CW"
    default_language = "pap"
    supported_languages = ("en_US", "nl", "pap", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # 1954: Creation of the Netherlands Antilles.
        if year <= 1953:
            return None

        super()._populate(year)

        # Aña Nobo.
        # Status: In-Use.

        # New Year's Day
        self._add_new_years_day(tr("Aña Nobo"))

        # Dialuna despues di Carnaval Grandi.
        # Status: In-Use.
        # Started in 1947.

        # Carnival Monday
        self._add_ash_monday(tr("Dialuna despues di Carnaval Grandi"))

        # Bièrnèsantu.
        # Status: In-Use.

        # Good Friday
        self._add_good_friday(tr("Bièrnèsantu"))

        # Pasku di Resurekshon.
        # Status: In-Use

        # Easter Sunday
        self._add_easter_sunday(tr("Pasku di Resurekshon"))

        # Di dos dia di Pasku di Resurekshon.
        # Status: In-Use.

        # Easter Monday
        self._add_easter_monday(tr("Di dos dia di Pasku di Resurekshon"))

        # Dia di la Reina/Dia di Rey.
        # Status: In-Use.
        # Started under Queen Wilhelmina in 1891.
        # Queen Beatrix kept Queen Juliana's Birthday after her coronation.
        # Switched to Aña di Rey in 2014 for King Willem-Alexander.
        # Have its name changed again to Dia di Rey from 2021 onwards.

        # King's / Queen's Day
        name = (
            # King's Day.
            tr("Dia di Rey")
            if year >= 2014
            # Queen's Day.
            else tr("Dia di la Reina")
        )
        if year >= 2014:
            dt = date(year, APR, 27)
        else:
            dt = date(year, APR, 30)
        if self._is_sunday(dt):
            dt += td(days=-1) if year >= 1980 else td(days=+1)
        self._add_holiday(name, dt)

        # Dia di Obrero.
        # Status: In-Use.
        # If fall on Sunday, then this will be move to next working day.

        dt = date(year, MAY, 1)
        if self._is_sunday(dt) or (self._is_monday(dt) and year <= 1979):
            dt += td(days=+1)
        # Labor Day
        self._add_holiday(tr("Dia di Obrero"), dt)

        # Dia di Asenshon.
        # Status: In-Use.

        # Ascension Day
        self._add_ascension_thursday(tr("Dia di Asenshon"))

        # Dia di Himno i Bandera.
        # Status: In-Use.
        # Starts in 1984.

        if year >= 1984:
            # National Anthem and Flag Day
            self._add_holiday(tr("Dia di Himno i Bandera"), JUL, 2)

        # Dia di Pais Kòrsou / Dia di autonomia.
        # Status: In-Use.
        # Starts in 2010.

        if year >= 2010:
            # Curaçao Day
            self._add_holiday(tr("Dia di Pais Kòrsou"), OCT, 10)

        # Pasku di Nasementu.
        # Status: In-Use.

        # Christmas Day
        self._add_christmas_day(tr("Pasku di Nasementu"))

        # Di dos dia di Pasku di Nasementu.
        # Status: In-Use.

        # Second Day of Christmas
        self._add_christmas_day_two(tr("Di dos dia di Pasku di Nasementu"))


class CW(Curacao):
    pass


class CUW(Curacao):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import FRI, SAT, JUN
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import IslamicHolidays, InternationalHolidays

# Since Djibouti share most of it's holidays with other muslim countries,
# this class is just a copy of Egypt's.


class Djibouti(HolidayBase, IslamicHolidays, InternationalHolidays):
    # Holidays here are estimates, it is common for the day to be pushed
    # if falls in a weekend, although not a rule that can be implemented.
    # Holidays after 2020: the following four moving date holidays whose exact
    # date is announced yearly are estimated (and so denoted):
    # - Eid El Fetr*
    # - Eid El Adha*
    # - Isra wal Miraj*
    # - Moulad El Naby*
    # - Arafat*
    # *only if hijri-converter library is installed, otherwise a warning is
    #  raised that this holiday is missing. hijri-converter requires
    #  Python >= 3.6
    # is_weekend function is there, however not activated for accuracy.

    country = "DJ"
    weekend = {FRI, SAT}

    def __init__(self, *args, **kwargs):
        IslamicHolidays.__init__(self)
        InternationalHolidays.__init__(self)

        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self._add_new_years_day("Nouvel an")

        # Labour Day
        self._add_labor_day("Fête du travail")

        # Fête de l'indépendance
        self._add_holiday("Fête de l'indépendance", JUN, 27)
        self._add_holiday("Fête de l'indépendance", JUN, 28)

        # Isra wal Miraj
        # The night journey of the prophet Muhammad
        self._add_isra_and_miraj_day("Isra wal Miraj")

        # Eid al-Fitr - Feast Festive
        self._add_eid_al_fitr_day("Eid al-Fitr")
        self._add_eid_al_fitr_day_two("Eid al-Fitr deuxième jour")

        # Arafat & Eid al-Adha - Scarfice Festive
        self._add_arafah_day("Arafat")
        self._add_eid_al_adha_day("Eid al-Adha")
        self._add_eid_al_adha_day_two("Eid al-Adha deuxième jour")

        # Islamic New Year - (hijari_year, 1, 1)
        self._add_islamic_new_year_day("Nouvel an musulman")

        # Prophet Muhammad's Birthday - (hijari_year, 3, 12)
        self._add_mawlid_day("Naissance du prophet Muhammad")


class DJ(Djibouti):
    pass


class DJI(Djibouti):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)
#  Copyright: Kateryna Golovanova <kate@kgthreads.com>, 2022

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars import _CustomIslamicCalendar
from holidays.calendars.gregorian import GREGORIAN_CALENDAR
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class BosniaAndHerzegovina(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Bosnia_and_Herzegovina
    https://www.paragraf.ba/neradni-dani-fbih.html
    https://www.paragraf.ba/neradni-dani-republike-srpske.html
    https://www.paragraf.ba/neradni-dani-brcko.html
    """

    country = "BA"
    default_language = "bs"
    supported_languages = ("bs", "en_US", "sr", "uk")
    subdivisions = (
        "BIH",  # Federacija Bosne i Hercegovine
        "BRC",  # Brčko distrikt
        "SRP",  # Republika Srpska
    )
    _deprecated_subdivisions = (
        "BD",
        "FBiH",
        "RS",
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=BosniaAndHerzegovinaIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _add_observed(
        self, dt: date, include_sat: bool = True, include_sun: bool = True, days: int = +1
    ) -> None:
        # BIH: if first day of New Year's Day and Labor Day fall on Sunday, observed on Tuesday.
        # BRC: if holiday fall on Sunday, observed on next working day.
        # SRP: if second day of New Year's Day and Labor Day fall on Sunday, observed on Monday.
        if not self.observed:
            return None
        if (include_sun and self._is_sunday(dt)) or (include_sat and self._is_saturday(dt)):
            self._add_holiday(
                self.tr("%s (preneseno)") % self[dt],
                dt + td(days=+2 if self._is_saturday(dt) else days),
            )

    def _populate(self, year):
        super()._populate(year)

        # Orthodox Good Friday.
        self._add_good_friday(tr("Veliki petak (Pravoslavni)"))

        # Catholic Easter Monday.
        self._add_easter_monday(tr("Uskrsni ponedjeljak (Katolički)"), GREGORIAN_CALENDAR)

        # Eid al-Fitr.
        self._add_eid_al_fitr_day(tr("Ramazanski Bajram"))

        # Eid al-Adha.
        self._add_eid_al_adha_day(tr("Kurban Bajram"))

        if self.subdiv == "BD":
            self._add_subdiv_brc_holidays()
        elif self.subdiv == "FBiH":
            self._add_subdiv_bih_holidays()
        elif self.subdiv == "RS":
            self._add_subdiv_srp_holidays()

    def _add_subdiv_holidays(self):
        if not self.subdiv:
            # New Year's Day.
            name = tr("Nova godina")
            self._add_new_years_day(name)
            self._add_new_years_day_two(name)

            # Orthodox Christmas.
            self._add_christmas_day(tr("Božić (Pravoslavni)"))

            # Labor Day.
            name = tr("Međunarodni praznik rada")
            self._add_labor_day(name)
            self._add_labor_day_two(name)

            # Catholic Christmas.
            self._add_christmas_day(tr("Božić (Katolički)"), GREGORIAN_CALENDAR)

        super()._add_subdiv_holidays()

    def _add_subdiv_bih_holidays(self):
        # New Year's Day.
        name = tr("Nova godina")
        self._add_observed(self._add_new_years_day(name), include_sat=False, days=+2)
        self._add_new_years_day_two(name)

        # Orthodox Christmas Eve.
        self._add_christmas_eve(tr("Badnji dan (Pravoslavni)"))

        # Orthodox Christmas.
        self._add_christmas_day(tr("Božić (Pravoslavni)"))

        # Independence Day.
        self._add_holiday(tr("Dan nezavisnosti"), MAR, 1)

        # Catholic Good Friday.
        self._add_good_friday(tr("Veliki petak (Katolički)"), GREGORIAN_CALENDAR)

        # Catholic Easter.
        self._add_easter_sunday(tr("Uskrs (Katolički)"), GREGORIAN_CALENDAR)

        # Orthodox Easter.
        self._add_easter_sunday(tr("Vaskrs (Pravoslavni)"))

        # Orthodox Easter Monday.
        self._add_easter_monday(tr("Uskrsni ponedjeljak (Pravoslavni)"))

        # Labor Day.
        name = tr("Međunarodni praznik rada")
        self._add_observed(self._add_labor_day(name), include_sat=False, days=+2)
        self._add_labor_day_two(name)

        # Victory Day.
        self._add_world_war_two_victory_day(tr("Dan pobjede nad fašizmom"))

        # Statehood Day.
        self._add_holiday(tr("Dan državnosti"), NOV, 25)

        # Catholic Christmas Eve.
        self._add_christmas_eve(tr("Badnji dan (Katolički)"), GREGORIAN_CALENDAR)

        # Catholic Christmas.
        self._add_christmas_day(tr("Božić (Katolički)"), GREGORIAN_CALENDAR)

        # Eid al-Fitr.
        self._add_eid_al_fitr_day_two(tr("Ramazanski Bajram"))

        # Eid al-Adha.
        self._add_eid_al_adha_day_two(tr("Kurban Bajram"))

    def _add_subdiv_brc_holidays(self):
        # New Year's Day.
        name = tr("Nova godina")
        self._add_observed(self._add_new_years_day(name), days=+2)
        self._add_new_years_day_two(name)

        # Orthodox Christmas.
        self._add_observed(self._add_christmas_day(tr("Božić (Pravoslavni)")), include_sat=False)

        self._add_observed(
            # Day of establishment of Brčko District.
            self._add_holiday(tr("Dan uspostavljanja Brčko distrikta"), MAR, 8),
            include_sat=False,
        )

        # Labor Day.
        name = tr("Međunarodni praznik rada")
        self._add_observed(self._add_labor_day(name), days=+2)
        self._add_labor_day_two(name)

        self._add_observed(
            # Catholic Christmas.
            self._add_christmas_day(tr("Božić (Katolički)"), GREGORIAN_CALENDAR),
            include_sat=False,
        )

    def _add_subdiv_srp_holidays(self):
        # New Year's Day.
        name = tr("Nova godina")
        self._add_observed(self._add_new_years_day(name), include_sun=False)
        self._add_new_years_day_two(name)

        # Orthodox Christmas Eve.
        self._add_christmas_eve(tr("Badnji dan (Pravoslavni)"))

        # Orthodox Christmas.
        self._add_christmas_day(tr("Božić (Pravoslavni)"))

        # Orthodox New Year.
        self._add_holiday(tr("Pravoslavna Nova godina"), JAN, 14)

        # Catholic Good Friday.
        self._add_good_friday(tr("Veliki petak (Katolički)"), GREGORIAN_CALENDAR)

        # Catholic Easter.
        self._add_easter_sunday(tr("Uskrs (Katolički)"), GREGORIAN_CALENDAR)

        # Orthodox Easter.
        self._add_easter_sunday(tr("Vaskrs (Pravoslavni)"))

        # Orthodox Easter Monday.
        self._add_easter_monday(tr("Uskrsni ponedjeljak (Pravoslavni)"))

        # Labor Day.
        name = tr("Međunarodni praznik rada")
        self._add_observed(self._add_labor_day(name), include_sun=False)
        self._add_labor_day_two(name)

        # Victory Day.
        self._add_world_war_two_victory_day(tr("Dan pobjede nad fašizmom"))

        self._add_holiday(
            # Dayton Agreement Day.
            tr("Dan uspostave Opšteg okvirnog sporazuma za mir u Bosni i Hercegovini"),
            NOV,
            21,
        )

        # Catholic Christmas Eve.
        self._add_christmas_eve(tr("Badnji dan (Katolički)"), GREGORIAN_CALENDAR)

        # Catholic Christmas.
        self._add_christmas_day(tr("Božić (Katolički)"), GREGORIAN_CALENDAR)

        # Eid al-Fitr.
        self._add_eid_al_fitr_day_two(tr("Ramazanski Bajram"))

        # Eid al-Adha.
        self._add_eid_al_adha_day_two(tr("Kurban Bajram"))


class BA(BosniaAndHerzegovina):
    pass


class BIH(BosniaAndHerzegovina):
    pass


class BosniaAndHerzegovinaIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2004: (FEB, 2),
        2005: (JAN, 21),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 9),
        2009: (NOV, 28),
        2010: (NOV, 17),
        2011: (NOV, 7),
        2012: (OCT, 26),
        2013: (OCT, 15),
        2014: (OCT, 4),
        2015: (SEP, 24),
        2016: (SEP, 13),
        2017: (SEP, 2),
        2018: (AUG, 22),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 9),
        2023: (JUN, 28),
    }

    EID_AL_FITR_DATES = {
        2001: (DEC, 17),
        2002: (DEC, 6),
        2003: (NOV, 26),
        2004: (NOV, 14),
        2005: (NOV, 4),
        2006: (OCT, 24),
        2007: (OCT, 13),
        2008: (OCT, 2),
        2009: (SEP, 21),
        2010: (SEP, 10),
        2011: (AUG, 31),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 28),
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2018: (JUN, 15),
        2019: (JUN, 4),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 21),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import GREGORIAN_CALENDAR, MAR, OCT, MON
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Greece(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Greece holidays.

    References:
     - https://en.wikipedia.org/wiki/Public_holidays_in_Greece
    """

    country = "GR"
    default_language = "el"
    supported_languages = ("el", "en_US")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Πρωτοχρονιά"))

        # Epiphany.
        self._add_epiphany_day(tr("Θεοφάνεια"), GREGORIAN_CALENDAR)

        # Clean Monday.
        self._add_ash_monday(tr("Καθαρά Δευτέρα"))

        # Independence Day.
        self._add_holiday(tr("Εικοστή Πέμπτη Μαρτίου"), MAR, 25)

        # Good Friday.
        self._add_good_friday(tr("Μεγάλη Παρασκευή"))

        # Easter Monday.
        self._add_easter_monday(tr("Δευτέρα του Πάσχα"))

        # Monday of the Holy Spirit.
        self._add_whit_monday(tr("Δευτέρα του Αγίου Πνεύματος"))

        # Labour Day.
        name = self.tr("Εργατική Πρωτομαγιά")
        name_observed = self.tr("%s (παρατηρήθηκε)")

        dt = self._add_labor_day(name)
        if self.observed and self._is_weekend(dt):
            # https://en.wikipedia.org/wiki/Public_holidays_in_Greece
            labour_day_observed_date = self._get_nth_weekday_from(1, MON, dt)
            # In 2016 and 2021, Labour Day coincided with other holidays
            # https://www.timeanddate.com/holidays/greece/labor-day
            if self.get(labour_day_observed_date):
                labour_day_observed_date += td(days=+1)
            self._add_holiday(name_observed % name, labour_day_observed_date)

        # Assumption of Mary.
        self._add_assumption_of_mary_day(tr("Κοίμηση της Θεοτόκου"))

        # Ochi Day.
        self._add_holiday(tr("Ημέρα του Όχι"), OCT, 28)

        # Christmas Day.
        self._add_christmas_day(tr("Χριστούγεννα"), GREGORIAN_CALENDAR)

        # Day after Christmas.
        self._add_christmas_day_two(tr("Επόμενη ημέρα των Χριστουγέννων"), GREGORIAN_CALENDAR)


class GR(Greece):
    pass


class GRC(Greece):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import JUL, AUG, NOV, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Colombia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Colombia has 18 holidays. The establishing of these are by:
    Ley 35 de 1939 (DEC 4): https://bit.ly/3PJwk7B
    Decreto 2663 de 1950 (AUG 5): https://bit.ly/3PJcut8
    Decreto 3743 de 1950 (DEC 20): https://bit.ly/3B9Otr3
    Ley 51 de 1983 (DEC 6): https://bit.ly/3aSobiB
    """

    country = "CO"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _move_holiday(self, dt: date) -> None:
        """
        On the 6th of December 1983, the government of Colombia declared which
        holidays are to take effect, and also clarified that a subset of them
        are to take place the next Monday if they do not fall on a Monday.
        This law is "Ley 51 de 1983" which translates to law 51 of 1983.
        Link: https://bit.ly/3PtPi2e
        A few links below to calendars from the 1980s to demonstrate this law
        change. In 1984 some calendars still use the old rules, presumably
        because they were printed prior to the declaration of law change.
        1981: https://bit.ly/3BbgKOc
        1982: https://bit.ly/3BdbhWW
        1984: https://bit.ly/3PqGxWU
        1984: https://bit.ly/3B7ogt8
        """

        if self.observed and not self._is_monday(dt) and self._year >= 1984:
            self._add_holiday(
                self.tr("%s (Observado)") % self[dt], self._get_nth_weekday_from(1, MON, dt)
            )
            self.pop(dt)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        if year >= 1951:
            # Epiphany.
            self._move_holiday(self._add_epiphany_day(tr("Día de los Reyes Magos")))

            # Saint Joseph's Day.
            self._move_holiday(self._add_saint_josephs_day(tr("Día de San José")))

            # Maundy Thursday.
            self._add_holy_thursday(tr("Jueves Santo"))

            # Good Friday.
            self._add_good_friday(tr("Viernes Santo"))

            # Ascension of Jesus.
            self._move_holiday(self._add_ascension_thursday(tr("Ascensión del señor")))

            # Corpus Christi.
            self._move_holiday(self._add_corpus_christi_day(tr("Corpus Christi")))

        # Labor Day.
        self._add_labor_day(tr("Día del Trabajo"))

        if year >= 1984:
            self._move_holiday(
                # Sacred Heart.
                self._add_holiday(tr("Sagrado Corazón"), self._easter_sunday + td(days=+68))
            )

        if year >= 1951:
            # Saint Peter and Saint Paul's Day.
            self._move_holiday(self._add_saints_peter_and_paul_day(tr("San Pedro y San Pablo")))

        # Independence Day.
        self._add_holiday(tr("Día de la Independencia"), JUL, 20)

        # Battle of Boyaca.
        self._add_holiday(tr("Batalla de Boyacá"), AUG, 7)

        if year >= 1951:
            # Assumption of Mary.
            self._move_holiday(self._add_assumption_of_mary_day(tr("La Asunción")))

        # Columbus Day.
        self._move_holiday(self._add_columbus_day(tr("Día de la Raza")))

        if year >= 1951:
            # All Saints' Day.
            self._move_holiday(self._add_all_saints_day(tr("Día de Todos los Santos")))

        self._move_holiday(
            # Independence of Cartagena.
            self._add_holiday(tr("Independencia de Cartagena"), NOV, 11)
        )

        if year >= 1951:
            # Immaculate Conception.
            self._add_immaculate_conception_day(tr("La Inmaculada Concepción"))

        # Christmas.
        self._add_christmas_day(tr("Navidad"))


class CO(Colombia):
    pass


class COL(Colombia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

# flake8: noqa: F401

from holidays.groups.buddhist import BuddhistCalendarHolidays
from holidays.groups.chinese import ChineseCalendarHolidays
from holidays.groups.christian import ChristianHolidays
from holidays.groups.hindu import HinduCalendarHolidays
from holidays.groups.international import InternationalHolidays
from holidays.groups.islamic import IslamicHolidays
from holidays.groups.thai import ThaiCalendarHolidays



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

# flake8: noqa: F401

from holidays.groups import (
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
    ThaiCalendarHolidays,
)



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, APR, MAY, JUN, AUG, SEP, OCT, NOV, DEC
from holidays.calendars.thai import KHMER_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import InternationalHolidays, ThaiCalendarHolidays


class Cambodia(HolidayBase, InternationalHolidays, ThaiCalendarHolidays):
    """
    A subclass of :py:class:`HolidayBase` representing public holidays in Cambodia.

    References:

    - Based on: https://www.nbc.gov.kh/english/news_and_events/official_holiday.php
                https://www.nbc.gov.kh/news_and_events/official_holiday.php
                https://en.wikipedia.org/wiki/Public_holidays_in_Cambodia

    - Checked with: https://asean.org/wp-content/uploads/2021/12/ASEAN-National-Holidays-2022.pdf
                    https://asean.org/wp-content/uploads/2022/12/ASEAN-Public-Holidays-2023.pdf
                    https://www.timeanddate.com/holidays/cambodia/

    Limitations:

    - Cambodian holidays only works from 1993 onwards.

    - Exact Public Holidays as per Cambodia's Official Gazette are only
      available from 2015 onwards.

    - Cambodian Lunar Calendar Holidays only work from 1941 (B.E. 2485) onwards until 2057
      (B.E. 2601) as we only have Thai year-type data for cross-checking until then.


    Country created by: `PPsyrius <https://github.com/PPsyrius>`__

    Country maintained by: `PPsyrius <https://github.com/PPsyrius>`__
    """

    country = "KH"
    default_language = "km"

    # Special Cases.

    sangkranta_in_lieu_covid = tr(
        # Khmer New Year's Replacement Holiday
        "ថ្ងៃឈប់សម្រាកសងជំនួសឲ្យពិធីបុណ្យចូលឆ្នាំថ្មីប្រពៃណីជាតិ"
    )
    # Special Public Holiday
    special_in_lieu_holidays = tr("ថ្ងៃឈប់សម្រាកសងជំនួស")

    special_holidays = {
        2016: (
            (MAY, 2, special_in_lieu_holidays),
            (MAY, 16, special_in_lieu_holidays),
        ),
        2018: (MAY, 21, special_in_lieu_holidays),
        2019: (SEP, 30, special_in_lieu_holidays),
        2020: (
            (MAY, 11, special_in_lieu_holidays),
            (AUG, 17, sangkranta_in_lieu_covid),
            (AUG, 18, sangkranta_in_lieu_covid),
            (AUG, 19, sangkranta_in_lieu_covid),
            (AUG, 20, sangkranta_in_lieu_covid),
            (AUG, 21, sangkranta_in_lieu_covid),
        ),
    }
    supported_languages = ("en_US", "km", "th")

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        ThaiCalendarHolidays.__init__(self, KHMER_CALENDAR)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # Available post-Independence from 1993 afterwards
        if year <= 1992:
            return None

        super()._populate(year)

        # Fixed Holidays

        #  ទិវាចូលឆ្នាំសាកល
        # Status: In-Use.

        # International New Year Day.
        self._add_new_years_day(tr("ទិវាចូលឆ្នាំសាកល"))

        #  ទិវាជ័យជម្នះលើរបបប្រល័យពូជសាសន៍
        # Status: In-Use.
        # Commemorates the end of the Khmer Rouge regime in 1979

        # Day of Victory over the Genocidal Regime.
        self._add_holiday(tr("ទិវាជ័យជម្នះលើរបបប្រល័យពូជសាសន៍"), JAN, 7)

        # ទិវាអន្តរជាតិនារី
        # Status: In-Use.

        # International Women's Rights Day
        self._add_womens_day(tr("ទិវាអន្តរជាតិនារី"))

        #  ពិធីបុណ្យចូលឆ្នាំថ្មីប្រពៃណីជាតិ
        # Status: In-Use.
        # Usually falls on April 13th except for 2017-2018 and 2021-2023 for years 2001-2050.

        if year != 2020:
            # Khmer New Year's Day
            sangkranta = tr("ពិធីបុណ្យចូលឆ្នាំថ្មីប្រពៃណីជាតិ")
            sangkranta_years_apr14 = {2017, 2018, 2021, 2022, 2023}
            dt = self._add_holiday(sangkranta, APR, 14 if year in sangkranta_years_apr14 else 13)
            self._add_holiday(sangkranta, dt + td(days=+1))
            self._add_holiday(sangkranta, dt + td(days=+2))

        #  ទិវាពលកម្មអន្តរជាតិ
        # Status: In-Use.

        # International Labor Day
        self._add_labor_day(tr("ទិវាពលកម្មអន្តរជាតិ"))

        # ព្រះរាជពិធីបុណ្យចម្រើនព្រះជន្ម ព្រះករុណា ព្រះបាទសម្តេចព្រះបរមនាថ នរោត្តម សីហមុនី
        # Status: In-Use.
        # Assumed to start in 2005. Was celebrated for 3 days until 2020.

        if year >= 2005:
            king_sihamoni_bday = tr(
                # Birthday of His Majesty Preah Bat Samdech Preah Boromneath
                # NORODOM SIHAMONI, King of Cambodia
                "ព្រះរាជពិធីបុណ្យចម្រើនព្រះជន្ម ព្រះករុណា ព្រះបាទសម្តេចព្រះបរមនាថ នរោត្តម សីហមុនី"
            )
            dt = self._add_holiday(king_sihamoni_bday, MAY, 14)
            if year <= 2019:
                self._add_holiday(king_sihamoni_bday, MAY, 13)
                self._add_holiday(king_sihamoni_bday, MAY, 15)

        # ទិវាជាតិនៃការចងចាំ
        # Status: Defunct.
        # Active between 2018-2019 as Public Holiday.
        # Was ទិវាចងកំហឹង (National Day of Anger) between 1983-2000.
        # Its celebration was put onhold by UN administration with
        # its name changed to present one in 2001.

        if 2018 <= year <= 2019:
            # National Day of Remembrance
            self._add_holiday(tr("ទិវាជាតិនៃការចងចាំ"), MAY, 20)

        # ទិវាកុមារអន្តរជាតិ
        # Status: Defunct.
        # Assumed to start in 1993, defunct from 2020 onwards.

        if year <= 2019:
            # International Children Day
            self._add_childrens_day(tr("ទិវាកុមារអន្តរជាតិ"))

        # ព្រះរាជពិធីបុណ្យចម្រើនព្រះជន្ម សម្តេចព្រះមហាក្សត្រី ព្រះវររាជមាតា នរោត្តម មុនិនាថ សីហនុ
        # Status: In-Use.
        # Assumed to start in 1994. A public holiday since 2015 at least.

        if year >= 1994:
            self._add_holiday(
                # Birthday of Her Majesty the Queen-Mother NORODOM MONINEATH SIHANOUK of Cambodia
                tr(
                    "ព្រះរាជពិធីបុណ្យចម្រើនព្រះជន្ម សម្តេចព្រះមហាក្សត្រី ព្រះវររាជមាតា នរោត្តម "
                    "មុនិនាថ សីហនុ"
                ),
                JUN,
                18,
            )

        # ទិវាប្រកាសរដ្ឋធម្មនុញ្ញ
        # Status: In-Use.
        # Starts in 1993

        # Constitution Day
        self._add_holiday(tr("ទិវាប្រកាសរដ្ឋធម្មនុញ្ញ"), SEP, 24)

        # ទិវាប្រារព្ឋពិធីគោរពព្រះវិញ្ញាណក្ខន្ឋ ព្រះករុណា ព្រះបាទសម្តេចព្រះ នរោត្តម សីហនុ
        # ព្រះមហាវីរក្សត្រ ព្រះវររាជបិតាឯករាជ្យ បូរណភាពទឹកដី និងឯកភាពជាតិខ្មែរ ព្រះបរមរតនកោដ្ឋ
        # Status: In-Use.
        # Starts in 2012.

        if year >= 2012:
            self._add_holiday(
                # Mourning Day of the Late King-Father
                # NORODOM SIHANOUK of Cambodia
                tr(
                    "ទិវាប្រារព្ឋពិធីគោរពព្រះវិញ្ញាណក្ខន្ឋ ព្រះករុណា ព្រះបាទសម្តេចព្រះ នរោត្តម "
                    "សីហនុ ព្រះមហាវីរក្សត្រ ព្រះវររាជបិតាឯករាជ្យ បូរណភាពទឹកដី និងឯកភាពជាតិខ្មែរ "
                    "ព្រះបរមរតនកោដ្ឋ"
                ),
                OCT,
                15,
            )

        # ទិវារំលឹកសន្ធិសញ្ញាសន្តិភាពទីក្រុងប៉ារីស
        # Status: Defunct.
        # Assumed to start in 1993, defunct from 2020 onwards.

        if year <= 2019:
            # Paris Peace Agreement's Day
            self._add_holiday(tr("ទិវារំលឹកសន្ធិសញ្ញាសន្តិភាពទីក្រុងប៉ារីស"), OCT, 23)

        # ព្រះរាជពិធីគ្រងព្រះបរមរាជសម្បត្តិ របស់ ព្រះករុណា
        # ព្រះបាទសម្តេចព្រះបរមនាថ នរោត្តម សីហមុនី
        # ព្រះមហាក្សត្រនៃព្រះរាជាណាចក្រកម្ពុជា
        # Status: In-Use.
        # Starts in 2004.

        if year >= 2004:
            self._add_holiday(
                # Coronation Day of His Majesty Preah Bat Samdech Preah
                # Boromneath NORODOM SIHAMONI, King of Cambodia
                tr(
                    "ព្រះរាជពិធីគ្រងព្រះបរមរាជសម្បត្តិ របស់ ព្រះករុណា "
                    "ព្រះបាទសម្តេចព្រះបរមនាថ នរោត្តម សីហមុនី "
                    "ព្រះមហាក្សត្រនៃព្រះរាជាណាចក្រកម្ពុជា"
                ),
                OCT,
                29,
            )

        # ពិធីបុណ្យឯករាជ្យជាតិ
        # Status: In-Use.
        # Starts in 1953

        # National Independence Day
        self._add_holiday(tr("ពិធីបុណ្យឯករាជ្យជាតិ"), NOV, 9)

        # ទិវាសិទ្ធិមនុស្សអន្តរជាតិ
        # Status: Defunct.
        # Assumed to start in 1993, defunct from 2020 onwards.

        if year <= 2019:
            # International Human Rights Day
            self._add_holiday(tr("ទិវាសិទ្ធិមនុស្សអន្តរជាតិ"), DEC, 10)

        # Cambodian Lunar Calendar Holidays
        # See `_ThaiLunisolar` in holidays/utils.py for more details.
        # Cambodian Lunar Calendar Holidays only work from 1941 to 2057.

        # ពិធីបុណ្យមាឃបូជា
        # Status: Defunct.
        # 15th Waxing Day of Month 3.
        # Defunct from 2020 onwards.

        if year <= 2019:
            # Meak Bochea Day
            self._add_makha_bucha(tr("ពិធីបុណ្យមាឃបូជា"))

        # ពិធីបុណ្យវិសាខបូជា
        # Status: In-Use.
        # 15th Waxing Day of Month 6.
        # This utilizes Thai calendar as a base, though are calculated to always happen
        # in the Traditional Visakhamas month (May).

        # Visaka Bochea Day
        self._add_visakha_bucha(tr("ពិធីបុណ្យវិសាខបូជា"))

        # ព្រះរាជពិធីច្រត់ព្រះនង្គ័ល
        # Status: In-Use.
        # 4th Waning Day of Month 6.
        # Unlike Thai ones, Cambodian Royal Ploughing Ceremony is always fixed.

        # Royal Ploughing Ceremony
        self._add_preah_neangkoal(tr("ព្រះរាជពិធីច្រត់ព្រះនង្គ័ល"))

        # ពិធីបុណ្យភ្ផុំបិណ្ឌ
        # Status: In-Use.
        # 14th Waning Day of Month 10 - 1st Waxing Day of Month 11.
        # The 3rd day is added as a public holiday from 2017 onwards.

        # Pchum Ben Day
        pchum_ben = tr("ពិធីបុណ្យភ្ផុំបិណ្ឌ")
        pchum_ben_date = self._add_pchum_ben(pchum_ben)
        if pchum_ben_date:
            self._add_holiday(pchum_ben, pchum_ben_date + td(days=-1))
            if year >= 2017:
                self._add_holiday(pchum_ben, pchum_ben_date + td(days=+1))

        # ព្រះរាជពិធីបុណ្យអុំទូក បណ្តែតប្រទីប និងសំពះព្រះខែអកអំបុក
        # Status: In-Use.
        # 14th Waxing Day of Month 12 - 1st Waning Day of Month 12.

        # Water Festival
        bon_om_touk = tr("ព្រះរាជពិធីបុណ្យអុំទូក បណ្តែតប្រទីប និងសំពះព្រះខែអកអំបុក")
        bon_om_touk_date = self._add_loy_krathong(bon_om_touk)
        if bon_om_touk_date:
            self._add_holiday(bon_om_touk, bon_om_touk_date + td(days=-1))
            self._add_holiday(bon_om_touk, bon_om_touk_date + td(days=+1))


class KH(Cambodia):
    pass


class KHM(Cambodia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, MAR, APR, JUL, AUG, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import InternationalHolidays, IslamicHolidays


class Tunisia(HolidayBase, InternationalHolidays, IslamicHolidays):
    """
    Holidays here are estimates, it is common for the day to be pushed
    if falls in a weekend, although not a rule that can be implemented.
    Holidays after 2020: the following four moving date holidays whose exact
    date is announced yearly are estimated (and so denoted):
    - Eid El Fetr
    - Eid El Adha
    - Arafat Day
    - Moulad El Naby
    """

    country = "TN"
    default_language = "ar"
    # Estimated label.
    estimated_label = tr("(تقدير*) *%s")
    supported_languages = ("ar", "en_US")

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self._add_new_years_day(tr("رأس السنة الميلادية"))

        # Revolution and Youth Day - January 14
        self._add_holiday(tr("عيد الثورة والشباب"), JAN, 14)

        # Independence Day
        self._add_holiday(tr("عيد الإستقلال"), MAR, 20)

        # Martyrs' Day
        self._add_holiday(tr("عيد الشهداء"), APR, 9)

        # Labour Day
        self._add_labor_day(tr("عيد العمال"))

        # Republic Day
        self._add_holiday(tr("عيد الجمهورية"), JUL, 25)

        # Women's Day
        self._add_holiday(tr("عيد المرأة"), AUG, 13)

        # Evacuation Day
        self._add_holiday(tr("عيد الجلاء"), OCT, 15)

        # Eid al-Fitr - Feast Festive
        # date of observance is announced yearly, This is an estimate since
        # having the Holiday on Weekend does change the number of days,
        # deceided to leave it since marking a Weekend as a holiday
        # wouldn't do much harm.
        name = tr("عيد الفطر")
        self._add_eid_al_fitr_day(name)
        # Eid al-Fitr Holiday
        self._add_eid_al_fitr_day_two(tr("عطلة عيد الفطر"))
        self._add_eid_al_fitr_day_three(tr("عطلة عيد الفطر"))

        #  Eid al-Adha - Scarfice Festive
        # date of observance is announced yearly
        name = tr("عيد الأضحى")
        # Arafat Day
        self._add_arafah_day(tr("يوم عرفة"))
        self._add_eid_al_adha_day(name)
        # Eid al-Adha Holiday
        self._add_eid_al_adha_day_two(tr("عطلة عيد الأضحى"))
        self._add_eid_al_adha_day_three(tr("عطلة عيد الأضحى"))

        # Islamic New Year - (hijari_year, 1, 1)
        self._add_islamic_new_year_day(tr("رأس السنة الهجرية"))

        # Prophet Muhammad's Birthday - (hijari_year, 3, 12)
        self._add_mawlid_day(tr("عيد المولد النبوي"))


class TN(Tunisia):
    pass


class TUN(Tunisia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

# flake8: noqa: F403

from holidays.constants import *
from holidays.holiday_base import *
from holidays.registry import EntityLoader
from holidays.utils import *

__version__ = "0.29"


EntityLoader.load("countries", globals())
EntityLoader.load("financial", globals())



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import FEB, APR, AUG, DEC, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Zimbabwe(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Zimbabwe
    https://en.wikipedia.org/wiki/Robert_Gabriel_Mugabe_National_Youth_Day
    """

    country = "ZW"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_sunday(dt):
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=days))

    def _populate(self, year):
        if year <= 1987:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_observed(self._add_new_years_day("New Year's Day"))

        if year >= 2018:
            self._add_observed(
                # Robert Gabriel Mugabe National Youth Day.
                self._add_holiday("Robert Gabriel Mugabe National Youth Day", FEB, 21)
            )

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Easter Saturday.
        self._add_holy_saturday("Easter Saturday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Independence Day.
        apr_18 = self._add_holiday("Independence Day", APR, 18)
        self._add_observed(apr_18, days=+2 if apr_18 == self._easter_sunday else +1)

        # Workers' Day.
        self._add_observed(self._add_labor_day("Workers' Day"))

        # Africa Day.
        self._add_observed(self._add_africa_day("Africa Day"))

        second_mon_of_august = self._add_holiday(
            # Zimbabwe Heroes' Day.
            "Zimbabwe Heroes' Day",
            self._get_nth_weekday_of_month(2, MON, AUG),
        )

        # Defense Forces Day.
        self._add_holiday("Defense Forces Day", second_mon_of_august + td(days=+1))

        # Unity Day.
        self._add_observed(self._add_holiday("Unity Day", DEC, 22))

        # Christmas Day.
        self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)

        # Boxing Day.
        self._add_observed(self._add_christmas_day_two("Boxing Day"))


class ZW(Zimbabwe):
    pass


class ZWE(Zimbabwe):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import MAR, MAY, SEP
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays, IslamicHolidays

# Ethiopian holidays are estimated: it is common for the day to be pushed
# if falls in a weekend, although not a rule that can be implemented.
# Holidays after 2020: the following four moving date holidays whose exact
# date is announced yearly are estimated (and so denoted):
# - Eid El Fetr*
# - Eid El Adha*
# - Arafat Day*
# - Moulad El Naby*
# *only if hijri-converter library is installed, otherwise a warning is
#  raised that this holiday is missing. hijri-converter requires
#  Python >= 3.6
# is_weekend function is there, however not activated for accuracy.


class Ethiopia(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    country = "ET"
    default_language = "am"
    estimated_label = tr("%s* (*ግምት)")
    supported_languages = ("am", "ar", "en_US")

    @staticmethod
    def _is_leap_year(year):
        """
        Ethiopian leap years are coincident with leap years in the Gregorian
        calendar until the end of February 2100. It starts earlier from new
        year of western calendar.
        Ethiopian leap year starts on Sep 11, so it has an effect on
        holidays between Sep 11 and Jan 1. Therefore, here on the following
        function we intentionally add 1 to the leap year to offset the
        difference.
        """
        return HolidayBase._is_leap_year(year + 1)

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        # The Ethiopian New Year is called Kudus Yohannes in Ge'ez and
        # Tigrinya, while in Amharic,
        # the official language of Ethiopia it is called Enkutatash.
        # It occurs on September 11 in the Gregorian Calendar;
        # except for the year preceding a leap year, when it occurs on
        # September 12.

        # Ethiopian New Year.
        self._add_holiday(tr("አዲስ ዓመት እንቁጣጣሽ"), SEP, 12 if self._is_leap_year(year) else 11)

        # Finding of True Cross.
        self._add_holiday(tr("መስቀል"), SEP, 28 if self._is_leap_year(year) else 27)

        # Orthodox Christmas.
        self._add_christmas_day(tr("ገና"))

        # Orthodox Epiphany.
        self._add_epiphany_day(tr("ጥምቀት"))

        # Orthodox Good Friday.
        self._add_good_friday(tr("ስቅለት"))

        # Orthodox Easter Sunday.
        self._add_easter_sunday(tr("ፋሲካ"))

        if year > 1896:
            # Adwa Victory Day.
            self._add_holiday(tr("አድዋ"), MAR, 2)

        # Labour Day.
        self._add_labor_day(tr("የሰራተኞች ቀን"))

        if year > 1941:
            # Patriots Day.
            self._add_holiday(tr("የአርበኞች ቀን"), MAY, 5)

        if year > 1991:
            # Downfall of Dergue Regime Day.
            self._add_holiday(tr("ደርግ የወደቀበት ቀን"), MAY, 28)

        if year < 1991 and year > 1974:
            # Downfall of King Haile Selassie.
            self._add_holiday(tr("ደርግ የመጣበት ቀን"), SEP, 13 if self._is_leap_year(year) else 12)

        # Eid al-Fitr.
        self._add_eid_al_fitr_day(tr("ኢድ አልፈጥር"))

        # Eid al-Adha.
        name = tr("አረፋ")
        self._add_eid_al_adha_day(name)
        self._add_eid_al_adha_day_two(name)

        # Prophet Muhammad's Birthday.
        name = tr("መውሊድ")
        self._add_mawlid_day(name)
        self._add_mawlid_day_two(name)


class ET(Ethiopia):
    pass


class ETH(Ethiopia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, MAR, JUN, OCT, DEC, FRI, SAT, SUN
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Sweden(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Swedish holidays.
    Note that holidays falling on a sunday are "lost",
    it will not be moved to another day to make up for the collision.
    In Sweden, ALL sundays are considered a holiday
    (https://sv.wikipedia.org/wiki/Helgdagar_i_Sverige).
    Initialize this class with include_sundays=False
    to not include sundays as a holiday.
    Primary sources:
    https://sv.wikipedia.org/wiki/Helgdagar_i_Sverige and
    http://www.riksdagen.se/sv/dokument-lagar/dokument/svensk-forfattningssamling/lag-1989253-om-allmanna-helgdagar_sfs-1989-253
    """

    country = "SE"
    default_language = "sv"
    supported_languages = ("en_US", "sv", "uk")

    def __init__(self, include_sundays=True, **kwargs):
        """
        :param include_sundays: Whether to consider sundays as a holiday
        (which they are in Sweden)
        :param kwargs:
        """
        self.include_sundays = include_sundays
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Nyårsdagen"))

        # Epiphany.
        self._add_epiphany_day(tr("Trettondedag jul"))

        if year <= 1953:
            # Feast of the Annunciation.
            self._add_holiday(tr("Jungfru Marie bebådelsedag"), MAR, 25)

        # Good Friday.
        self._add_good_friday(tr("Långfredagen"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Påskdagen"))

        # Easter Monday.
        self._add_easter_monday(tr("Annandag påsk"))

        # Source: https://sv.wikipedia.org/wiki/F%C3%B6rsta_maj
        if year >= 1939:
            # May Day.
            self._add_labor_day(tr("Första maj"))

        # Ascension Day.
        self._add_ascension_thursday(tr("Kristi himmelsfärdsdag"))

        # Source: https://sv.wikipedia.org/wiki/Sveriges_nationaldag
        if year >= 2005:
            # National Day of Sweden.
            self._add_holiday(tr("Sveriges nationaldag"), JUN, 6)

        # Whit Sunday.
        self._add_whit_sunday(tr("Pingstdagen"))

        if year <= 2004:
            # Whit Monday.
            self._add_whit_monday(tr("Annandag pingst"))

        # Source:
        # https://sv.wikipedia.org/wiki/Midsommarafton
        # https://www.nordiskamuseet.se/aretsdagar/midsommarafton
        # Midsummer evening. Friday between June 19th and June 25th
        dt = self._get_nth_weekday_from(1, FRI, JUN, 19) if year >= 1953 else date(year, JUN, 23)
        # Midsummer Eve.
        self._add_holiday(tr("Midsommarafton"), dt)

        # Midsummer Day.
        self._add_holiday(tr("Midsommardagen"), dt + td(days=+1))

        # All Saints' Day.
        self._add_holiday(tr("Alla helgons dag"), self._get_nth_weekday_from(1, SAT, OCT, 31))

        # Christmas Eve.
        self._add_christmas_eve(tr("Julafton"))

        # Christmas Day.
        self._add_christmas_day(tr("Juldagen"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Annandag jul"))

        # New Year's Eve.
        self._add_new_years_eve(tr("Nyårsafton"))

        if self.include_sundays:
            # Optionally add all Sundays of the year.
            begin = self._get_nth_weekday_of_month(1, SUN, JAN)
            end = date(year, DEC, 31)
            for dt in (begin + td(days=n) for n in range(0, (end - begin).days + 1, 7)):
                # Sunday.
                self._add_holiday(tr("Söndag"), dt)


class SE(Sweden):
    pass


class SWE(Sweden):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import FEB, MAR, MAY, JUN, AUG, OCT, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Ireland(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Official holidays in Ireland, as declared in the Citizen's Information
    bulletin:
    https://www.citizensinformation.ie/en/employment/employment_rights_and_conditions/leave_and_holidays/public_holidays_in_ireland.html
    """

    country = "IE"
    special_holidays = {
        2022: (MAR, 18, "Day of Remembrance and Recognition"),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_weekend(dt):
            self._add_holiday(
                "%s (Observed)" % self[dt], dt + td(days=+2 if self._is_saturday(dt) else days)
            )

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # St. Brigid's Day.
        if year >= 2023:
            dt = date(year, FEB, 1)
            self._add_holiday(
                "St. Brigid's Day",
                dt if self._is_friday(dt) else self._get_nth_weekday_from(1, MON, dt),
            )

        # St. Patrick's Day.
        self._add_observed(self._add_holiday("St. Patrick's Day", MAR, 17))

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # May Day.
        if year >= 1978:
            self._add_holiday(
                "May Day",
                date(year, MAY, 8)
                if year == 1995
                else self._get_nth_weekday_of_month(1, MON, MAY),
            )

        # June Bank holiday.
        self._add_holiday("June Bank Holiday", self._get_nth_weekday_of_month(1, MON, JUN))

        # Summer Bank holiday.
        self._add_holiday("August Bank Holiday", self._get_nth_weekday_of_month(1, MON, AUG))

        # October Bank Holiday.
        self._add_holiday("October Bank Holiday", self._get_nth_weekday_of_month(-1, MON, OCT))

        # Christmas Day.
        self._add_observed(self._add_christmas_day("Christmas Day"))

        # St. Stephen's Day.
        self._add_observed(self._add_christmas_day_two("St. Stephen's Day"), days=+2)


class IE(Ireland):
    pass


class IRL(Ireland):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import MAY, SEP, AUG, OCT, DEC
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class NorthMacedonia(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_North_Macedonia
    """

    country = "MK"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        self._add_new_years_day("New Year's Day")

        self._add_christmas_day("Christmas Day (Orthodox)")

        self._add_easter_monday("Easter Monday (Orthodox)")

        self._add_labor_day("Labour Day")

        self._add_holiday("Saints Cyril and Methodius Day", MAY, 24)

        self._add_holiday("Republic Day", AUG, 2)

        self._add_holiday("Independence Day", SEP, 8)

        self._add_holiday("Day of Macedonian Uprising in 1941", OCT, 11)

        self._add_holiday("Day of the Macedonian Revolutionary Struggle", OCT, 23)

        self._add_holiday("Saint Clement of Ohrid Day", DEC, 8)

        self._add_eid_al_fitr_day("Eid al-Fitr")


class MK(NorthMacedonia):
    pass


class MKD(NorthMacedonia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import OCT, NOV
from holidays.constants import BANK, PUBLIC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Austria(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "AT"
    default_language = "de"
    supported_categories = {BANK, PUBLIC}
    supported_languages = ("de", "en_US", "uk")
    subdivisions = ("1", "2", "3", "4", "5", "6", "7", "8", "9")

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        if isinstance(kwargs.get("subdiv"), int):
            kwargs["subdiv"] = str(kwargs["subdiv"])
        # Default subdivision to "9"
        if not kwargs.get("subdiv", kwargs.get("state")):
            kwargs["subdiv"] = "9"
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # New Year's Day.
        self._add_new_years_day(tr("Neujahr"))

        # Epiphany.
        self._add_epiphany_day(tr("Heilige Drei Könige"))

        # Easter Monday.
        self._add_easter_monday(tr("Ostermontag"))

        # Labor Day.
        self._add_labor_day(tr("Staatsfeiertag"))

        # Ascension Day.
        self._add_ascension_thursday(tr("Christi Himmelfahrt"))

        # Whit Monday.
        self._add_whit_monday(tr("Pfingstmontag"))

        # Corpus Christi.
        self._add_corpus_christi_day(tr("Fronleichnam"))

        # Assumption Day.
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))

        if 1919 <= self._year <= 1934:
            # National Day.
            self._add_holiday(tr("Nationalfeiertag"), NOV, 12)
        if self._year >= 1967:
            self._add_holiday(tr("Nationalfeiertag"), OCT, 26)

        # All Saints' Day.
        self._add_all_saints_day(tr("Allerheiligen"))

        # Immaculate Conception.
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

        # Christmas Day.
        self._add_christmas_day(tr("Christtag"))

        # St. Stephen's Day.
        self._add_christmas_day_two(tr("Stefanitag"))

    def _populate_bank_holidays(self):
        # Good Friday.
        self._add_good_friday(tr("Karfreitag"))

        # Christmas Eve.
        self._add_christmas_eve(tr("Heiliger Abend"))

        # New Year's Eve.
        self._add_new_years_eve(tr("Silvester"))


class AT(Austria):
    pass


class AUT(Austria):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Portugal(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    A subclass of :py:class:`HolidayBase` representing public holidays
    in Portugal.

    References:

    - Based on:
        https://en.wikipedia.org/wiki/Public_holidays_in_Portugal

    National Level:
    - [Labour Day]
        https://www.e-konomista.pt/dia-do-trabalhador/
    - [Portugal Day]
        Decreto 17.171
    - [Restoration of Independence Day]
        Gazeta de Lisboa, 8 de Dezembro de 1823 (n.º 290), pp. 1789 e 1790

    Regional Level:
    - [Azores]
        https://files.dre.pt/1s/1980/08/19200/23052305.pdf
    - [Madeira]
        https://files.dre.pt/1s/1979/11/25900/28782878.pdf
        https://files.dre.pt/1s/1989/02/02800/04360436.pdf
        https://files.dre.pt/1s/2002/11/258a00/71837183.pdf

    """

    country = "PT"
    default_language = "pt_PT"

    # https://en.wikipedia.org/wiki/ISO_3166-2:PT
    # `Ext` represents the national holidays most people have off
    subdivisions = (
        "01",
        "02",
        "03",
        "04",
        "05",
        "06",
        "07",
        "08",
        "09",
        "10",
        "11",
        "12",
        "13",
        "14",
        "15",
        "16",
        "17",
        "18",
        "20",
        "30",
        "Ext",
    )
    supported_languages = ("en_US", "pt_PT")

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        self._add_new_years_day(tr("Ano Novo"))

        # Carnival is no longer a holiday, but some companies let workers off.
        # TODO: recollect the years in which it was a public holiday
        # self[e + td(days=-47)] = "Carnaval"

        # Good Friday.
        self._add_good_friday(tr("Sexta-feira Santa"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Páscoa"))

        # Revoked holidays in 2013–2015.
        if year <= 2012 or year >= 2016:
            self._add_corpus_christi_day(tr("Corpo de Deus"))
            if year >= 1910:
                self._add_holiday(tr("Implantação da República"), OCT, 5)
            self._add_all_saints_day(tr("Dia de Todos os Santos"))
            if year >= 1823:
                self._add_holiday(tr("Restauração da Independência"), DEC, 1)

        if year >= 1974:
            self._add_holiday(tr("Dia da Liberdade"), APR, 25)
            self._add_labor_day(tr("Dia do Trabalhador"))
        if year >= 1911:
            if 1933 <= year <= 1973:
                self._add_holiday(tr("Dia de Camões, de Portugal e da Raça"), JUN, 10)
            elif year >= 1978:
                self._add_holiday(
                    tr("Dia de Portugal, de Camões e das Comunidades Portuguesas"), JUN, 10
                )
            else:
                self._add_holiday(tr("Dia de Portugal"), JUN, 10)

        self._add_assumption_of_mary_day(tr("Assunção de Nossa Senhora"))
        self._add_immaculate_conception_day(tr("Imaculada Conceição"))
        self._add_christmas_day(tr("Dia de Natal"))

    def _add_subdiv_holidays(self):
        if self._year >= 1911:
            super()._add_subdiv_holidays()

    def _add_subdiv_ext_holidays(self):
        """
        Adds extended days that most people have as a bonus from their
        companies:

        - Carnival
        - the day before and after xmas
        - the day before the new year
        - Lisbon's city holiday
        """

        # TODO: add bridging days:
        # - get Holidays that occur on Tuesday  and add Monday (-1 day)
        # - get Holidays that occur on Thursday and add Friday (+1 day)

        self._add_carnival_monday(tr("Carnaval"))
        self._add_christmas_eve(tr("Véspera de Natal"))
        self._add_christmas_day_two(tr("26 de Dezembro"))
        self._add_new_years_eve(tr("Véspera de Ano Novo"))
        self._add_holiday(tr("Dia de Santo António"), JUN, 13)

    def _add_subdiv_01_holidays(self):
        self._add_holiday(tr("Dia de Santa Joana"), MAY, 12)

    def _add_subdiv_02_holidays(self):
        self._add_ascension_thursday(tr("Quinta-feira da Ascensão"))

    def _add_subdiv_03_holidays(self):
        self._add_holiday(tr("Dia de São João"), JUN, 24)

    def _add_subdiv_04_holidays(self):
        self._add_holiday(tr("Dia de Nossa Senhora das Graças"), AUG, 22)

    def _add_subdiv_05_holidays(self):
        self._add_holiday(
            tr("Dia de Nossa Senhora de Mércoles"), self._easter_sunday + td(days=+16)
        )

    def _add_subdiv_06_holidays(self):
        self._add_holiday(tr("Dia de Santa Isabel"), JUL, 4)

    def _add_subdiv_07_holidays(self):
        self._add_holiday(tr("Dia de São Pedro"), JUN, 29)

    def _add_subdiv_08_holidays(self):
        self._add_holiday(tr("Dia do Município de Faro"), SEP, 7)

    def _add_subdiv_09_holidays(self):
        self._add_holiday(tr("Dia do Município da Guarda"), NOV, 27)

    def _add_subdiv_10_holidays(self):
        self._add_holiday(tr("Dia do Município de Leiria"), MAY, 22)

    def _add_subdiv_11_holidays(self):
        self._add_holiday(tr("Dia de Santo António"), JUN, 13)

    def _add_subdiv_12_holidays(self):
        self._add_holiday(tr("Dia do Município de Portalegre"), MAY, 23)

    def _add_subdiv_13_holidays(self):
        self._add_holiday(tr("Dia de São João"), JUN, 24)

    def _add_subdiv_14_holidays(self):
        self._add_saint_josephs_day(tr("Dia de São José"))

    def _add_subdiv_15_holidays(self):
        self._add_holiday(tr("Dia de Bocage"), SEP, 15)

    def _add_subdiv_16_holidays(self):
        self._add_holiday(tr("Dia de Nossa Senhora da Agonia"), AUG, 20)

    def _add_subdiv_17_holidays(self):
        self._add_holiday(tr("Dia de Santo António"), JUN, 13)

    def _add_subdiv_18_holidays(self):
        self._add_holiday(tr("Dia de São Mateus"), SEP, 21)

    def _add_subdiv_20_holidays(self):
        if self._year >= 1981:
            self._add_holiday(
                tr("Dia da Região Autónoma dos Açores"), self._easter_sunday + td(days=+50)
            )

    def _add_subdiv_30_holidays(self):
        if 1979 <= self._year <= 1988:
            self._add_holiday(tr("Dia da Região Autónoma da Madeira"), JUL, 1)
        elif self._year >= 1989:
            self._add_holiday(
                tr("Dia da Região Autónoma da Madeira e das Comunidades Madeirenses"), JUL, 1
            )

        if self._year >= 2002:
            self._add_holiday(tr("Primeira Oitava"), DEC, 26)


class PT(Portugal):
    pass


class PRT(Portugal):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import APR, MAY, AUG
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Netherlands(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_the_Netherlands
    http://www.iamsterdam.com/en/plan-your-trip/practical-info/public-holidays
    """

    country = "NL"
    default_language = "nl"
    supported_languages = ("en_US", "nl", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Nieuwjaarsdag"))

        # Good Friday.
        self._add_good_friday(tr("Goede Vrijdag"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Eerste paasdag"))

        # Easter Monday.
        self._add_easter_monday(tr("Tweede paasdag"))

        # King's / Queen's day
        if year >= 1891:
            name = (
                # King's Day.
                tr("Koningsdag")
                if year >= 2014
                # Queen's Day.
                else tr("Koninginnedag")
            )
            if year >= 2014:
                dt = date(year, APR, 27)
            elif year >= 1949:
                dt = date(year, APR, 30)
            else:
                dt = date(year, AUG, 31)
            if self._is_sunday(dt):
                dt += td(days=-1) if year >= 1980 else td(days=+1)
            self._add_holiday(name, dt)

        if year >= 1945 and year % 5 == 0:
            # Liberation Day.
            self._add_holiday(tr("Bevrijdingsdag"), MAY, 5)

        # Ascension Day.
        self._add_ascension_thursday(tr("Hemelvaartsdag"))

        # Whit Sunday.
        self._add_whit_sunday(tr("Eerste Pinksterdag"))

        # Whit Monday.
        self._add_whit_monday(tr("Tweede Pinksterdag"))

        # Christmas Day.
        self._add_christmas_day(tr("Eerste Kerstdag"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Tweede Kerstdag"))


class NL(Netherlands):
    pass


class NLD(Netherlands):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import JUL
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Belgium(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Belgium
    https://www.belgium.be/nl/over_belgie/land/belgie_in_een_notendop/feestdagen
    https://nl.wikipedia.org/wiki/Feestdagen_in_Belgi%C3%AB
    """

    country = "BE"
    default_language = "nl"
    supported_languages = ("de", "en_US", "fr", "nl", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Nieuwjaar"))

        # Easter.
        self._add_easter_sunday(tr("Pasen"))

        # Easter Monday.
        self._add_easter_monday(tr("Paasmaandag"))

        # Labor Day.
        self._add_labor_day(tr("Dag van de Arbeid"))

        # Ascension Day.
        self._add_ascension_thursday(tr("Hemelvaart"))

        # Whit Sunday.
        self._add_whit_sunday(tr("Pinksteren"))

        # Whit Monday.
        self._add_whit_monday(tr("Pinkstermaandag"))

        # National Day.
        self._add_holiday(tr("Nationale feestdag"), JUL, 21)

        # Assumption of Mary.
        self._add_assumption_of_mary_day(tr("Onze Lieve Vrouw hemelvaart"))

        # All Saints' Day.
        self._add_all_saints_day(tr("Allerheiligen"))

        # Armistice Day.
        self._add_remembrance_day(tr("Wapenstilstand"))

        # Christmas Day.
        self._add_christmas_day(tr("Kerstmis"))


class BE(Belgium):
    pass


class BEL(Belgium):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars import (
    _CustomBuddhistCalendar,
    _CustomChineseCalendar,
    _CustomIslamicCalendar,
)
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import (
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    InternationalHolidays,
    IslamicHolidays,
)


class Indonesia(
    HolidayBase,
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    InternationalHolidays,
    IslamicHolidays,
):
    """
    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_Indonesia
    - https://www.timeanddate.com/holidays/indonesia
    - https://www.officeholidays.com/countries/indonesia
    """

    country = "ID"
    special_holidays = {
        # Election Day.
        2018: (JUN, 27, "Hari Pemilihan"),
        2019: (APR, 17, "Hari Pemilihan"),
        2020: (DEC, 9, "Hari Pemilihan"),
    }

    def __init__(self, *args, **kwargs):
        BuddhistCalendarHolidays.__init__(
            self, calendar=IndonesiaBuddhistCalendar(), show_estimated=True
        )
        ChineseCalendarHolidays.__init__(
            self, calendar=IndonesiaChineseCalendar(), show_estimated=True
        )
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=IndonesiaIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("Tahun Baru Masehi")

        if year >= 2003:
            # Chinese New Year.
            self._add_chinese_new_years_day("Tahun Baru Imlek")

        if year >= 1983:
            # https://en.wikipedia.org/wiki/Nyepi
            dates_obs = {
                2009: (MAR, 26),
                2010: (MAR, 16),
                2011: (MAR, 5),
                2012: (MAR, 23),
                2013: (MAR, 12),
                2014: (MAR, 31),
                2015: (MAR, 21),
                2016: (MAR, 9),
                2017: (MAR, 28),
                2018: (MAR, 17),
                2019: (MAR, 7),
                2020: (MAR, 25),
                2021: (MAR, 14),
                2022: (MAR, 3),
                2023: (MAR, 22),
                2024: (MAR, 11),
                2025: (MAR, 29),
                2026: (MAR, 19),
            }
            if year in dates_obs:
                # Day of Silence.
                self._add_holiday("Hari Suci Nyepi", *dates_obs[year])

        # Eid al-Fitr.
        self._add_eid_al_fitr_day("Hari Raya Idul Fitri")
        self._add_eid_al_fitr_day_two("Hari kedua dari Hari Raya Idul Fitri")

        # Eid al-Adha.
        self._add_eid_al_adha_day("Hari Raya Idul Adha")

        # Islamic New Year.
        self._add_islamic_new_year_day("Tahun Baru Islam")

        # The Prophet's Birthday.
        self._add_mawlid_day("Maulid Nabi Muhammad")

        # The Prophet's Ascension.
        self._add_isra_and_miraj_day("Isra' Mi'raj Nabi Muhammad")

        # Good Friday.
        self._add_good_friday("Wafat Yesus Kristus")

        if year >= 1983:
            # Buddha's Birthday.
            self._add_vesak("Hari Raya Waisak")

        if 1953 <= year <= 1968 or year >= 2014:
            # Labor Day.
            self._add_labor_day("Hari Buruh Internasional")

        # Ascension Day.
        self._add_ascension_thursday("Kenaikan Yesus Kristus")

        if year >= 2017:
            # Pancasila Day.
            self._add_holiday("Hari Lahir Pancasila", JUN, 1)

        # Independence Day.
        self._add_holiday("Hari Kemerdekaan Republik Indonesia", AUG, 17)

        # Christmas Day.
        self._add_christmas_day("Hari Raya Natal")


class ID(Indonesia):
    pass


class IDN(Indonesia):
    pass


class IndonesiaBuddhistCalendar(_CustomBuddhistCalendar):
    VESAK_DATES = {
        2007: (JUN, 1),
        2008: (MAY, 20),
        2009: (MAY, 9),
        2010: (MAY, 28),
        2011: (MAY, 17),
        2012: (MAY, 6),
        2013: (MAY, 25),
        2014: (MAY, 15),
        2015: (JUN, 2),
        2016: (MAY, 22),
        2017: (MAY, 11),
        2018: (MAY, 29),
        2019: (MAY, 19),
        2020: (MAY, 7),
        2021: (MAY, 26),
        2022: (MAY, 16),
        2023: (JUN, 4),
    }


class IndonesiaChineseCalendar(_CustomChineseCalendar):
    LUNAR_NEW_YEAR_DATES = {
        2003: (FEB, 1),
        2004: (JAN, 22),
        2005: (FEB, 9),
        2006: (JAN, 30),
        2007: (FEB, 19),
        2008: (FEB, 7),
        2009: (JAN, 26),
        2010: (FEB, 15),
        2011: (FEB, 3),
        2012: (JAN, 23),
        2013: (FEB, 11),
        2014: (JAN, 31),
        2015: (FEB, 19),
        2016: (FEB, 8),
        2017: (JAN, 28),
        2018: (FEB, 16),
        2019: (FEB, 5),
        2020: (JAN, 25),
        2021: (FEB, 12),
        2022: (FEB, 1),
        2023: (JAN, 22),
    }


class IndonesiaIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2004: (FEB, 2),
        2005: (JAN, 21),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 8),
        2009: (NOV, 27),
        2010: (NOV, 17),
        2011: (NOV, 6),
        2012: (OCT, 26),
        2013: (OCT, 15),
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 12),
        2017: (SEP, 1),
        2018: (AUG, 22),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 10),
        2023: (JUN, 29),
    }

    EID_AL_FITR_DATES = {
        2001: (DEC, 16),
        2002: (DEC, 6),
        2003: (NOV, 25),
        2004: (NOV, 14),
        2005: (NOV, 3),
        2006: (OCT, 24),
        2007: (OCT, 13),
        2008: (OCT, 1),
        2009: (SEP, 20),
        2010: (SEP, 10),
        2011: (AUG, 30),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 28),
        2015: (JUL, 17),
        2016: (JUL, 6),
        2017: (JUN, 25),
        2018: (JUN, 15),
        2019: (JUN, 5),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 22),
    }

    HIJRI_NEW_YEAR_DATES = {
        2001: (MAR, 26),
        2002: (MAR, 15),
        2003: (MAR, 5),
        2004: (FEB, 22),
        2005: (FEB, 10),
        2006: (JAN, 31),
        2007: (JAN, 20),
        2008: ((JAN, 10), (DEC, 29)),
        2009: (DEC, 18),
        2010: (DEC, 7),
        2011: (NOV, 27),
        2012: (NOV, 15),
        2013: (NOV, 5),
        2014: (OCT, 25),
        2015: (OCT, 14),
        2016: (OCT, 2),
        2017: (SEP, 21),
        2018: (SEP, 11),
        2019: (SEP, 1),
        2020: (AUG, 20),
        2021: (AUG, 11),
        2022: (JUL, 30),
    }

    ISRA_AND_MIRAJ_DATES = {
        2001: (OCT, 15),
        2002: (OCT, 4),
        2003: (SEP, 24),
        2004: (SEP, 12),
        2005: (SEP, 1),
        2006: (AUG, 22),
        2007: (AUG, 11),
        2008: (JUL, 31),
        2009: (JUL, 20),
        2010: (JUL, 9),
        2011: (JUN, 29),
        2012: (JUN, 17),
        2013: (JUN, 6),
        2014: (MAY, 27),
        2015: (MAY, 16),
        2016: (MAY, 6),
        2017: (APR, 24),
        2018: (APR, 14),
        2019: (APR, 3),
        2020: (MAR, 22),
        2021: (MAR, 11),
        2022: (FEB, 28),
    }

    MAWLID_DATES = {
        2006: (APR, 10),
        2007: (MAR, 31),
        2008: (MAR, 20),
        2009: (MAR, 9),
        2010: (FEB, 26),
        2011: (FEB, 15),
        2012: (FEB, 5),
        2013: (JAN, 24),
        2014: (JAN, 14),
        2015: ((JAN, 3), (DEC, 24)),
        2016: (DEC, 12),
        2017: (DEC, 1),
        2018: (NOV, 20),
        2019: (NOV, 9),
        2020: (OCT, 29),
        2021: (OCT, 19),
        2022: (OCT, 8),
    }
